#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import sys, os, re, shlex, shutil, glob, subprocess, collections
from os import path
from datetime import datetime
import numpy as np
from scipy import interpolate
import matplotlib as mpl
from . import six


# Test afni installation
has_afni = bool(re.search('version', subprocess.check_output(['afni', '-ver']).decode('utf-8'), re.IGNORECASE))
# # Find afni path
# config_dir = path.expanduser('~/.mripy')
# if not path.exists(config_dir):
#     os.makedirs(config_dir)
# if has_afni:
#     config_file = path.join(config_dir, 'afni_path')
#     if path.exists(config_file):
#         with open(config_file, 'r') as f:
#             afni_path = f.readline()
#     else:
#         afni_path = subprocess.check_output('find ~ -iregex ".*/abin"', shell=True).decode('utf-8').split('\n')[0]
#         with open(config_file, 'w') as f:
#             f.write(afni_path)
# else:
#     afni_path = ''


def filter_output(lines, tags=None, pattern=None):
    '''
    Filter output lines according to their initial tags (++, *+, **, etc.) and/or
    a regex search pattern.

    Parameters
    ----------
    tags : list of tags
        Default is [], which means all lines will pass the filter.
    pattern : str
    '''
    if tags is None:
        tags = []
    if len(tags) > 0:
        lines = [line for line in lines if line[:2] in tags]
    if pattern is not None:
        lines = [line for line in lines if re.search(pattern, line)]
    return lines


def check_output(cmd, tags=None, pattern=None, verbose=0, **kwargs):
    '''
    The syntax of subprocess.check_output(shell=False) is tedious for long cmd.
    But for security reason, we don't want to use shell=True for external cmd.
    This helper function allows you to execute a single cmd without shell=True.

    Parameters
    ----------
    cmd : str
        A single command string packed with all options (but no wildcard)
    **kwargs :
        Go to subprocess.check_output(**kwargs)

    Returns
    -------
    lines : list of lines
        Much easier to deal with compared with subprocess.check_output()
    '''
    if isinstance(cmd, six.string_types):
        cmd = shlex.split(cmd) # Split by space, preserving quoted substrings
    lines = subprocess.check_output(cmd, stderr=subprocess.STDOUT, **kwargs).decode('utf-8').split('\n')
    lines = filter_output(lines, tags, pattern)
    if verbose:
        for line in lines:
            print(line, file=sys.stderr if line.startswith('*') else sys.stdout)
    return lines


def call(cmd):
    if isinstance(cmd, six.string_types):
        cmd = shlex.split(cmd) # Split by space, preserving quoted substrings
    cmd_str = ' '.join(cmd)
    print('>>', cmd_str)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=False)
    for line in iter(p.stdout.readline, b''): # The 2nd argument is sentinel character
        print(line.decode('utf-8'), end='')
    p.stdout.close() # Notify the child process that the PIPE has been broken
    if p.wait():
        raise RuntimeError(f'Error occurs when executing the following command (returncode={p.returncode}):\n{cmd_str}') 


def split_out_file(out_file, split_path=False, trailing_slash=False):
    '''
    Ensure that path.join(out_dir, prefix, ext) can be checked by path.exists().

    >>> split_out_file('dset.nii')
    ('dset', '.nii')
    >>> split_out_file('dset.1D')
    ('dset', '.1D')
    >>> split_out_file('folder/dset')
    ('folder/dset', '+orig.HEAD')
    >>> split_out_file('folder/dset+orig', split_path=True)
    ('folder', 'dset', '+orig.HEAD')
    >>> split_out_file('dset+orig.', split_path=True)
    ('', 'dset', '+orig.HEAD')
    >>> split_out_file('folder/dset+orig.HEAD', split_path=True, trailing_slash=True)
    ('folder/', 'dset', '+orig.HEAD')
    >>> split_out_file('dset+tlrc.BRIK', split_path=True, trailing_slash=True)
    ('', 'dset', '+tlrc.HEAD')
    '''
    out_dir, out_name = path.split(out_file)
    if trailing_slash and out_dir:
        out_dir += '/'
    match = re.match(r'(.+)(.nii|.nii.gz|.1D|.1D.dset|.1D.roi|.niml.dset|.niml.roi|.gii.dset|.csv)$', out_name)
    if match:
        prefix, ext = match.groups()
    else:
        match = re.match(r'(.+)(\+(?:orig|tlrc))(?:.|.HEAD|.BRIK)?$', out_name)
        if match:
            prefix, ext = match.groups()
            ext += '.HEAD'
        else:
            prefix = out_name
            ext = '+orig.HEAD'
    if split_path:
        return out_dir, prefix, ext
    else:
        return path.join(out_dir, prefix), ext


def insert_suffix(fname, suffix):
    prefix, ext = split_out_file(fname)
    return f"{prefix}{suffix}{ext}"


def get_prefix(fname, with_path=False):
    '''
    Return "dset" given "path/to/dset+orig.HEAD", "dset+orig.", "dset+tlrc", "dsets"
    '''
    if path.splitext(fname)[1] in ['.niml', '.1D', '.dset']: # For surface dset
        match = re.match(r'(.+)\.(?:niml|1D)(?:\.dset)?', fname)
        prefix = match.group(1)
    else: # For 3d dset
        # fstem = path.splitext(path.basename(fname))[0]
        if fname[-5:].upper() in ['.HEAD', '.BRIK']:
            fstem = fname[:-5]
        elif fname.endswith('.'):
            fstem = fname[:-1]
        else:
            fstem = fname
        prefix = fstem[:-5] if len(fstem) > 5 and fstem[-5:] in ['+orig', '+tlrc'] else fstem
    if not with_path:
        prefix = path.basename(prefix)
    return prefix


def get_surf_vol(suma_dir):
    '''
    Infer SUMA SurfVol filename with full path (agnostic about file type: .nii vs +orig.HEAD/BRIK).
    '''
    # TODO: SurfVol.depth.nii
    candidates = glob.glob(path.join(suma_dir, '*_SurfVol*'))
    candidates = [f for f in candidates if re.search(r'_SurfVol(?:\.nii|\+orig\.HEAD)', f)]
    if len(candidates) == 0:
        raise ValueError(f'>> Cannot identify SurfVol in "{suma_dir}"')
    else:
        return candidates[0]


def get_suma_subj(suma_dir):
    '''Infer SUMA subject given path to SUMA folder.'''
    match = re.match('(.+)_SurfVol.+', path.basename(get_surf_vol(suma_dir)))
    if match:
        return match.group(1)
    else:
        raise RuntimeError(f'>> Cannot infer SUMA subject from "{suma_dir}"')


def get_surf_type(suma_dir):
    '''Infer SUMA surface mesh file type (.gii vs .asc).'''
    surf_files = [f for f in os.listdir(suma_dir) if re.match('(?:lh|rh).(?:pial|smoothwm|inflated).*', f)]
    return path.splitext(surf_files[0])[1]


SPEC_HEMIS = ['lh', 'rh', 'both', 'mh', 'bh']
HEMI_PATTERN = r'(?:(?<=[^a-zA-Z0-9])|^)(?:lh|rh|both|mh|bh)(?=[^a-zA-Z0-9])'

def substitute_hemi(fname, hemi='{0}'):
    return re.sub(HEMI_PATTERN, hemi, fname)


def get_suma_spec(suma_spec):
    '''
    Infer other spec files from one spec file (either lh.spec, rh.spec, or both.spec).
    
    Parameters
    ----------
    suma_spec : str
        Either a .spec file or the suma_dir.
    '''
    if path.isdir(suma_spec): # It is actually the `suma_dir`
        subj = get_suma_subj(suma_spec)
        return {hemi: path.join(suma_spec, f"{subj}_{hemi}.spec") for hemi in SPEC_HEMIS}
    else: # It is a .spec file
        spec_fmt = re.sub(f"({'|'.join(SPEC_HEMIS)}).spec", '{0}.spec', suma_spec)
        return {hemi: spec_fmt.format(hemi) for hemi in SPEC_HEMIS}


def get_suma_info(suma_dir, suma_spec=None):
    info = {}
    info['subject'] = get_suma_subj(suma_dir)
    if suma_spec is None: # Infer spec files from suma_dir
        info['spec'] = get_suma_spec(suma_dir)
    else: # Infer other spec files from one spec file
        info['spec'] = get_suma_spec(suma_spec)
    return info


def get_hemi(fname):
    basename = path.basename(fname)
    match = re.search(HEMI_PATTERN, basename)
    if match:
        hemi = match.group(0)
    else:
        raise ValueError(f'** ERROR: Cannot infer "hemi" from "{basename}"')
    return hemi


def infer_surf_dset_variants(fname, hemis=SPEC_HEMIS):
    '''
    >>> infer_surf_dset_variants('data.niml.dset')
    {'lh': 'lh.data.niml.dset', 'rh': 'rh.data.niml.dset', 'both': 'both.data.niml.dset', mh': 'mh.data.niml.dset'}
    >>> infer_surf_dset_variants('lh.data.niml.dset')
    {'lh': 'lh.data.niml.dset'}

    Parameters
    ----------
    fname : str, list, or dict
    '''
    if isinstance(fname, six.string_types):
        match = re.search(HEMI_PATTERN, path.basename(fname))
        if match:
            fname = {match.group(0): fname}
        else:
            out_dir, prefix, ext = split_out_file(fname, split_path=True, trailing_slash=True)
            fname = {hemi: f"{out_dir}{hemi}.{prefix}{ext}" for hemi in hemis}
    if not isinstance(fname, dict):
        fdict = {}
        for f in fname:
            match = re.search(HEMI_PATTERN, path.basename(f))
            if match:
                fdict[match.group(0)] = f
            else:
                raise ValueError(f'** ERROR: Cannot infer "hemi" from "{path.basename(f)}"')
        fname = fdict
    return fname


def get_ORIENT(fname, format='str'):
    '''
    Parameters
    ----------
    format : str, {'code', 'str', 'mat', 'sorter'}

    References
    ----------
    [1] https://afni.nimh.nih.gov/pub/dist/doc/program_help/README.attributes.html
        #define ORI_R2L_TYPE  0  /* Right to Left         */
        #define ORI_L2R_TYPE  1  /* Left to Right         */
        #define ORI_P2A_TYPE  2  /* Posterior to Anterior */
        #define ORI_A2P_TYPE  3  /* Anterior to Posterior */
        #define ORI_I2S_TYPE  4  /* Inferior to Superior  */
        #define ORI_S2I_TYPE  5  /* Superior to Inferior  */

        Thus "0 3 4" is standard DICOM Reference Coordinates System, i.e., RAI.
        The AFNI convention is also that R-L, A-P, and I-S are negative-to-positive, i.e., RAI.

    [2] https://nipy.org/nibabel/nifti_images.html
        On the other hand, NIFTI images have an affine relating the voxel coordinates 
        to world coordinates in RAS+ space, or LPI in AFNI's term.
    '''
    res = check_output(['3dAttribute', 'ORIENT_SPECIFIC', fname])[-2]
    ORIENT = np.fromiter(map(int, res.split()), int)
    code2str = np.array(['R', 'L', 'P', 'A', 'I', 'S'])
    code2mat = np.array([[ 1, 0, 0],
                         [-1, 0, 0],
                         [ 0,-1, 0],
                         [ 0, 1, 0],
                         [ 0, 0, 1],
                         [ 0, 0,-1]])
    code2axis = np.array([0, 0, 1, 1, 2, 2])
    if format == 'code':
        return ORIENT
    elif format == 'str':
        return ''.join(code2str[ORIENT])
    elif format == 'mat':
        return code2mat[ORIENT]
    elif format == 'sorter':
        return np.argsort(code2axis[ORIENT])


def get_DIMENSION(fname):
    '''
    [x, y, z, t, 0]
    '''
    res = check_output(['3dAttribute', 'DATASET_DIMENSIONS', fname])[-2]
    DIMENSION = np.fromiter(map(int, res.split()), int)
    return DIMENSION


def get_ORIGIN(fname):
    res = check_output(['3dAttribute', 'ORIGIN', fname])[-2]
    ORIGIN = np.fromiter(map(float, res.split()), float)
    return ORIGIN


def get_DELTA(fname):
    res = check_output(['3dAttribute', 'DELTA', fname])[-2]
    DELTA = np.fromiter(map(float, res.split()), float)
    return DELTA


def get_affine(fname):
    ORIENT = get_ORIENT(fname, format='sorter')
    ORIGIN = get_ORIGIN(fname)
    DELTA = get_DELTA(fname)
    MAT = np.c_[np.diag(DELTA), ORIGIN][ORIENT,:]
    return MAT


def get_affine_nifti(fname):
    MAT = np.diag([-1,-1, 1]) @ get_affine(fname)
    return MAT


def get_dims(fname):
    '''
    Dimensions (number of voxels) of the data matrix.
    See also: get_head_dims
    '''
    # res = check_output(['@GetAfniDims', fname])[-2] # There can be leading warnings for oblique datasets
    res = check_output(['3dinfo', '-n4', fname])[-2] # `@GetAfniDims` may not work for things like `dset.nii'[0..10]'`
    return np.int_(res.split()) # np.fromiter(map(int, res.split()), int)


def get_head_dims(fname):
    '''
    Dimensions (number of voxels) along R-L, A-P, I-S axes.
    See also: get_dims
    '''
    res = check_output(['3dinfo', '-orient', '-n4', fname])[-2]
    res = res.split()
    orient = res[0]
    dims = np.int_(res[1:])
    ori2ax = {'R': 0, 'L': 0, 'A': 1, 'P': 1, 'I': 2, 'S': 2}
    axes = [ori2ax[ori] for ori in orient]
    return np.r_[dims[np.argsort(axes)], dims[3]]


def get_head_delta(fname):
    '''
    Resolution (voxel size) along R-L, A-P, I-S axes.
    '''
    res = check_output(['3dinfo', '-orient', '-d3', fname])[-2]
    res = res.split()
    orient = res[0]
    delta = np.abs(np.float_(res[1:]))
    ori2ax = {'R': 0, 'L': 0, 'A': 1, 'P': 1, 'I': 2, 'S': 2}
    axes = [ori2ax[ori] for ori in orient]
    return delta[np.argsort(axes)]


def get_head_extents(fname):
    '''
    Spatial extent along R, L, A, P, I and S.
    '''
    res = check_output(['3dinfo', '-extent', fname])[-2]
    return np.float_(res.split())


def get_brick_labels(fname, label2index=False):
    res = check_output(['3dAttribute', 'BRICK_LABS', fname])[-2]
    labels = res.split('~')[:-1] # Each label ends with "~"
    if label2index:
        return {label: k for k, label in enumerate(labels)}
    else:
        return np.array(labels)


def set_brick_labels(fname, labels):
    check_output(['3drefit', '-relabel_all_str', ' '.join(labels), fname])


def get_TR(fname):
    return float(check_output(['3dinfo', '-TR', fname])[-2])


def get_attribute(fname, name, type=None):
    res = check_output(['3dAttribute', name, fname])[-2]
    if type == 'int':
        return np.int_(res[:-1].split())
    elif type == 'float':
        return np.float_(res[:-1].split())
    else:
        return res[:-1]


def set_attribute(fname, name, value, type=None):
    values = np.atleast_1d(value)
    if type == 'str' or isinstance(value, str):
        check_output(['3drefit', '-atrstring', name, f"{value}", fname])
    elif type == 'int' or np.issubdtype(values.dtype, np.integer):
        check_output(['3drefit', '-atrint', name, f"{' '.join([str(v) for v in values])}", fname])
    elif type == 'float' or np.issubdtype(values.dtype, np.floating):
        check_output(['3drefit', '-atrfloat', name, f"{' '.join([str(v) for v in values])}", fname])


def get_nifti_field(fname, name, type=None):
    res = check_output(['nifti_tool', '-disp_hdr', '-field', name, '-infiles', fname])[-2]
    if type == 'int':
        return np.int_(res.split()[3:])
    elif type == 'float':
        return np.float_(res.split()[3:])
    else:
        return res[37:]


def set_nifti_field(fname, name, value, out_file=None):
    values = np.atleast_1d(value)
    check_output(['nifti_tool', '-mod_hdr', '-mod_field', name, f"{' '.join([str(v) for v in values])}", '-infiles', fname] 
        + (['-overwrite'] if out_file is None else ['-prefix', out_file]))


def get_S2E_mat(fname, mat='S2E'):
    mat = {'S2E': 'S2B', 'S2B': 'S2B', 'E2S': 'B2S', 'B2S': 'B2S'}[mat]
    res = check_output("cat_matvec -ONELINE '{0}::ALLINEATE_MATVEC_{1}_000000'".format(fname, mat))[-2]
    return np.float_(res.split()).reshape(3,4)


def generate_spec(fname, surfs, ext=None, **kwargs):
    if ext is None:
        ext = '.gii'
    defaults = dict(dict(type={'.asc': 'FS', '.gii': 'GII'}[ext], state=None, anat=None, parent=None), **kwargs)
    surfs = [dict(defaults, **({'name': surf} if isinstance(surf, six.string_types) else surf)) for surf in surfs]
    has_smoothwm = np.any([('smoothwm' in surf['name']) for surf in surfs])
    is_both = np.any([('lh' in surf['name']) for surf in surfs]) and np.any([('rh' in surf['name']) for surf in surfs])
    for surf in surfs:
        match = re.search(rf'([l|r]h)\.(.+)\.{ext[1:]}', surf['name'])
        surf['hemi'] = match.group(1)
        surf['surf'] = match.group(2)
        is_anat = surf['surf'] in ['pial', 'smoothwm', 'white']
        if surf['state'] is None:
            if not is_anat and is_both:
                surf['state'] = '_'.join([surf['surf'], surf['hemi']])
            else:
                surf['state'] = surf['surf']
        if surf['anat'] is None:
            surf['anat'] = 'Y' if is_anat else 'N'
        if surf['parent'] is None:
            if surf['name'] == 'smoothwm' or not has_smoothwm:
                surf['parent'] = 'SAME'
            else:
                surf['parent'] = '.'.join([surf['hemi'], 'smoothwm', ext[1:]])
    cmds = []
    for surf in surfs:
         cmds.extend(['-tsnad', surf['type'], surf['state'], surf['name'], surf['anat'], surf['parent']])
    subprocess.check_call(['quickspec', '-spec', fname, '-overwrite'] + cmds)


def update_afnirc(**kwargs):
    rc_file = path.expanduser('~/.afnirc')
    bak_file = path.expanduser('~/.afnirc.{0}.bak'.format(datetime.now().strftime('%Y%m%d')))
    if not path.exists(bak_file):
        shutil.copy(rc_file, bak_file)
    with open(rc_file, 'r') as fin:
        lines = fin.read().splitlines()
    updated = []
    is_managed = False
    managed_begin = '// Managed by mripy: begin'
    managed_end = '// Managed by mripy: end'
    managed = collections.OrderedDict()
    for line in lines:
        if not is_managed:
            if line == managed_begin:
                is_managed = True
            else:
                updated.append(line)
        else:
            if line == managed_end:
                is_managed = False
            else:
                match = re.search('(\S+)\s+=\s+((?:.(?!//))+)(?:\s+//\s+(.+))?', line)
                managed[match.group(1)] = (match.group(2).strip(), match.group(3)) # key, value, comment (can be None)
    for k, v in kwargs.items():
        if not isinstance(v, tuple):
            kwargs[k] = (v, None)
    managed.update(kwargs)
    n_managed = len([v for v in managed.values() if v[0] is not None])
    if n_managed > 0:
        if updated[-1] != '':
            updated.append('')
        updated.append(managed_begin)
    for key, (value, comment) in managed.items():
        if value is not None:
            updated.append('   {0: <24} = {1}'.format(key, value) +
                ('\t// {0}'.format(comment) if comment is not None else ''))
    if n_managed > 0:
        updated.append(managed_end)
    with open(rc_file, 'w') as fout:
        fout.write('\n'.join(updated))


def add_colormap(cmap, name=None, cyclic=False, index=None, categorical=False):
    '''
    cmap : list of RGB colors | matplotlib.colors.LinearSegmentedColormap
    '''
    if name is None:
        if isinstance(cmap, mpl.colors.LinearSegmentedColormap):
            name = cmap.name
        else:
            name = 'User{0:02d}'.format(index)
    if isinstance(cmap, mpl.colors.LinearSegmentedColormap):
        cmap = plots.get_color_list(cmap)
    if index is None:
        index = 1
    # Make colormap dir
    cmap_dir = path.expanduser('~/abin/colormaps')
    if not path.exists(cmap_dir):
        os.makedirs(cmap_dir)
    # Generate palette file
    temp_file = 'colors.tmp'
    with open(temp_file, 'w') as fout:
        fout.writelines(['\t'.join(map(str, color))+'\n' for color in cmap])
    cmap_file = path.join(cmap_dir, '{0}.pal'.format(name))
    with open(cmap_file, 'w') as fout:
        if categorical:
            subprocess.check_call(['MakeColorMap', '-f', temp_file, '-ah', name, '-nc', str(len(cmap))], stdout=fout)
        else:
            subprocess.check_call(['MakeColorMap', '-f', temp_file, '-ah', name] +
                (['-nc', str(128), '-sl'] if cyclic else ['-nc', str(129)]), stdout=fout)
    os.remove(temp_file)
    # Update .afnirc
    update_afnirc(**{'AFNI_COLORSCALE_{0:02d}'.format(index): path.relpath(cmap_file, path.expanduser('~'))})


def write_colorscale_file(fname, pal_name, colors, locations=None, interp=None):
    '''
    Parameters
    ----------
    fname : *.pal file name
    pal_name : palette name (or title)
    colors : a list of RGB colors within [0,1]
        first color (bottom) -> last color (top)
    locations : locations of the breakpoints where colors are defined
        0 (bottom) -> 1 (top)
    interp : 'linear'|'nearest'

    AFNI document says "There are exactly 128 color locations on an AFNI colorscale."
    For details, see https://afni.nimh.nih.gov/pub/dist/doc/OLD/afni_colorscale.html
    But in fact, if you fill the colorscale file with a lot of colors, only the first 256 colors will be used.
    '''
    if locations is None:
        locations = np.linspace(0, 1, len(colors))
    if interp is None:
        interp = 'linear'
    cmap = interpolate.interp1d(locations, colors, kind=interp, axis=0, bounds_error=False, fill_value='extrapolate')
    clist = [mpl.colors.to_hex(color) for color in cmap(np.linspace(0, 1, 256))]
    with open(fname, 'w') as fout:
        fout.write(f"{pal_name}\n")
        fout.writelines([f"{color}\n" for color in reversed(clist)])


def parse_patch(patch):
    '''
    Notes
    -----
    1. Each replacement is started with one or more comment lines. The last
       comment line is treated as replacement target, which may contain an
       optional replacement directive at the end:
        # This is an example <replace command="1"/>
       Possible directives for replacing the original scripts includes:
        1) command="n": replace n commands
        2) line="n": replace n lines
        3) until="regexp": replace until a specific line (the regexp is the
           last line to be replaced)
    2. Each replacement must end with two consecutive newlines.
    '''
    with open(patch, 'r') as fin:
        lines = fin.read().splitlines()
    replacements = []
    is_content = False
    n_newlines = 0
    for k, line in enumerate(lines):
        if is_content:
            contents.append(line)
            if line.strip() == '':
                n_newlines += 1
                if n_newlines >= 2:
                    is_content = False
            else:
                n_newlines = 0
            if not is_content or k+1 == len(lines):
                for kk in range(min(2, len(contents))):
                    if contents[-1] == '':
                        contents.pop(-1)
                    else:
                        break
                contents.append('# </patch>')
                replacements.append({'target': target, 'directives': directives, 'contents': contents})
        elif line[0] == '#':
            if k+1 < len(lines) and lines[k+1][0] != '#':
                match = re.match('((?:(?!<replace).)*)(?:<replace(.*)/>)?', line)
                target = match.group(1).rstrip()
                if match.group(2) is not None:
                    attributes = shlex.split(match.group(2).strip())
                    directives = dict([attr.split('=') for attr in attributes])
                else:
                    directives = {'command': 1}
                is_content = True
                contents = ['# <patch>']
    return replacements


def patch_afni_proc(original, patch, inplace=True):
    replacements = parse_patch(patch)
    n = 0
    with open(original, 'r') as fin:
        lines = fin.read().splitlines()
    patched = []
    n_to_replace = 0
    for k, line in enumerate(lines):
        if n == len(replacements):
            patched.append(line)
        else:
            replacement = replacements[n]
        if not n_to_replace:
            patched.append(line)
            match = re.search(replacement['target'], line)
            if match:
                replacement['indent'] = match.start()
                replacement['n_lines'] = six.MAXSIZE
                directives = replacement['directives']
                if 'command' in directives:
                    nc = 0
                    n_lines = 0
                    while nc < int(directives['command']):
                        n_lines += 1
                        x = lines[k+n_lines].strip()
                        if x != '' and x[0] != '#' and x[-1] != '\\':
                            nc += 1
                    replacement['n_lines'] = min(replacement['n_lines'], n_lines)
                if 'until' in directives:
                    n_lines = 0
                    while not re.match(directives['until'], lines[k+n_lines]):
                        n_lines += 1
                    replacement['n_lines'] = min(replacement['n_lines'], n_lines)
                if 'line' in directives:
                    replacement['n_lines'] = min(replacement['n_lines'], int(directives['line']))
                n_to_replace = replacement['n_lines']
        else:
            patched.append('# ' + line)
            n_to_replace -= 1
            if n_to_replace == 0:
                for content in replacement['contents']:
                    patched.append(' '*replacement['indent'] + content)
    if not inplace:
        p, f = path.split(original)
        fname = path.join(p, 'patched.'+f)
    else:
        shutil.copy(original, original+'.bak')
        fname = original
    with open(fname, 'w') as fout:
        fout.write('\n'.join(patched))


if __name__ == '__main__':
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
from os import path
from collections import OrderedDict
import numpy as np
from scipy import stats
import pandas as pd
from sklearn import model_selection, metrics
from . import six, utils


def standardize_within_group(X, groups, with_mean=True, with_std=True):
    '''
    This is an extension of the "mean centering" method proposed in [1].
    Can be used as a replacement for the training-set-wise standardization.
    Both with_mean and with_std may provide some extra performance.

    References
    ----------
    [1] Lee, S., & Kable, J. W. (2018). Simple but robust improvement in 
        multivoxel pattern classification. PloS One, 13(11), e0207083.
    '''
    X = X.copy()
    for g in np.unique(groups):
        indexer = (groups==g)
        u = X[indexer].mean(axis=0, keepdims=True)
        s = X[indexer].std(axis=0, keepdims=True) if with_std else 1
        if with_mean:
            X[indexer] = (X[indexer] - u) / s
        else:
            X[indexer] = (X[indexer] - u) / s + u
    return X


def permute_within_group(y, groups):
    y = y.copy()
    for g in np.unique(groups):
        indexer = (groups==g)
        y[indexer] = y[indexer][np.random.permutation(np.sum(indexer))]
    return y


def cross_validate_ext(model, X, y, groups=None, cv=None, pred_kws=None, method=None):
    if cv is None:
        cv = model_selection.LeaveOneGroupOut() # One group of each run
    if method is None:
        method = 'predict'
    pred_kws = dict(dict(), **({} if pred_kws is None else pred_kws))
    res = []
    idx = []
    for train_index, test_index in cv.split(X, y, groups):
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]
        model.fit(X_train, y_train)
        res.append(getattr(model, method)(X_test, **pred_kws))
        idx.extend(test_index)
    sorter = np.argsort(idx)
    if isinstance(res[0], tuple) and len(res[0]) > 1: # predict() has more the one output
        res = tuple([np.concatenate([r[k] for r in res], axis=0)[sorter] for k in range(len(res[0]))])
    else: # predict() has only one output
        res = np.concatenate(res, axis=0)[sorter]
    return res


def cross_validate_with_permutation(model, X, y, groups, rois=None, n_permutations=1000, scoring=None, cv=None):
    if rois is None:
        X, y, groups, rois = [X], [y], [groups], ['NA']
    if cv is None:
        cv = model_selection.LeaveOneGroupOut() # One group of each run
    if scoring is None:
        scoring = {'performance': metrics.make_scorer(metrics.accuracy_score)}
    def cross_validate(X, y, groups, roi, permute):
        if permute:
            y = permute_within_group(y, groups)
        scores = model_selection.cross_validate(model, X, y, groups, \
            scoring=scoring, cv=cv, return_train_score=True, n_jobs=1)
        res = OrderedDict(roi=roi, permute=permute, train=np.mean(scores['train_performance']), 
            test=np.mean(scores['test_performance']))
        return res
    res = []
    for XX, yy, gg, roi in zip(X, y, groups, rois):
        for permute in range(n_permutations+1):
            res.append(cross_validate(XX, yy, gg, roi, permute))
    res = pd.DataFrame(res)
    return res


def compute_critical_value(x, y, permute='permute', data=None, alpha=0.05, tail=2):
    '''
    Get critical values based on permutation distribution, and
    account for multiple comparisons using extreme statistics.

    Parameters
    ----------
    x : str, list of str
        Columns along which multiple comparisons occur (e.g., roi, time).
    y : str
        Column for performance measurement (e.g., test_accuracy, PC, RT).
    data : pd.DataFrame(x, y, permute)
        permute == 0 is originally observed data, >= 1 is permutation data.
    '''
    # Mean performance for each condition and each permutation
    by = [x, permute] if isinstance(x, six.string_types) else list(x) + [permute]
    df = data[data[permute]>0].groupby(by=by)[y].mean() # This is a Series with MultiIndex
    # Globally corrected critical value
    max_dist = df.groupby(permute).max().values # Max distribution
    min_dist = df.groupby(permute).min().values # Min distribution
    if tail == 2:
        gmax = np.percentile(max_dist, (1-alpha/2)*100)
        gmin = np.percentile(min_dist, alpha/2*100)
    else:
        gmax = np.percentile(max_dist, (1-alpha)*100)
        gmin = np.percentile(min_dist, alpha*100)
    # Per-comparison (uncorrected) critical value
    if tail == 2:
        pmax = df.groupby(x).quantile(1-alpha/2)
        pmin = df.groupby(x).quantile(alpha/2)
    else:
        pmax = df.groupby(x).quantile(1-alpha)
        pmin = df.groupby(x).quantile(alpha)
    bounds = pd.concat([pmin, pmax], axis=1)
    bounds.columns = ['lower', 'upper']
    bounds = pd.concat([pd.DataFrame([{'lower': gmin, 'upper': gmax}], index=['overall']), bounds], axis=0)
    # Determine significance
    obs = data[data[permute]==0]
    if obs.size > 0: # Contain originally observed data
        bounds['obs_mean'] = pd.concat([pd.Series([np.nan], ['overall']), obs.groupby(by=x)[y].mean()], axis=0) # Mean response
        bounds['obs_std'] = pd.concat([pd.Series([np.nan], ['overall']), obs.groupby(by=x)[y].std()], axis=0)
        bounds['obs_n'] = pd.concat([pd.Series([-1], ['overall']), obs.groupby(by=x)[y].count()], axis=0)
        n_comparisons = len(obs[x].unique())
        if tail == 2: # The two-tailed p value is twice the one-tailed p value (assuming you correctly predicted the direction of the difference)
            bounds['corrected'] = (bounds.obs_mean < bounds.lower['overall']) | (bounds.upper['overall'] < bounds.obs_mean) # Significance (corrected)
            bounds['uncorrected'] = (bounds.obs_mean < bounds.lower) | (bounds.upper < bounds.obs_mean) # Significance (uncorrected)
        elif tail == 1:
            bounds['corrected'] = (bounds.upper['overall'] < bounds.obs_mean)
            bounds['p_corr'] = [np.nan] + [1-stats.percentileofscore(max_dist, v)/100 for v in bounds.obs_mean[1:]]
            bounds['uncorrected'] = (bounds.upper < bounds.obs_mean)
            bounds['p_uncorr'] = [np.nan] + [1-stats.percentileofscore(df[k], v)/100 for k, v in bounds.obs_mean[1:].iteritems()]
            bounds['bonferroni'] = bounds['p_uncorr'] * n_comparisons
        elif tail == -1:
            bounds['corrected'] = (bounds.obs_mean < bounds.lower['overall'])
            bounds['uncorrected'] = (bounds.obs_mean < bounds.lower)
    return bounds


# def plot_permutation(x, y, subject='subject', permute='permute', data=None, plot=None, 
#     color=None, x_order=None, xtick_format=None, chance=0, alpha=0.05, tail=2, mcc='extreme', 
#     figsize=None, star_shift=None, star_alpha=None,
#     dist_kws=None, scatter_kws=None, line_kws=None, bar_kws=None, point_kws=None, chance_kws=None, ref_kws=None, 
#     show_mcc=True, show_n=True, show_p=False, show_ref=False, show_num=False):
#     '''
#     Parameters
#     ----------
#     x : str
#     y : str
#     data : pd.DataFrame(x, y, permute)
#         permute == 0 is originally observed data, >= 1 is permutation data.
#     '''
#     if figsize is not None:
#         fig = plt.figure(figsize=figsize)
#     else:
#         fig = plt.gcf()
#     pmt = data[data[permute]>0]
#     has_pmt = (pmt.size > 0)
#     obs = data[data[permute]==0]
#     if x_order is None:
#         x_order = data[x].unique()
#     else:
#         x_order = [x_label for x_label in x_order if x_label in data[x].values]
#     if has_pmt:
#         bounds = compute_critical_value(x=x, y=y, data=data, alpha=alpha, tail=tail)
#         df_pmt = pmt.groupby(by=[x, permute])[y].mean().reset_index()
#     df_obs = obs.groupby(by=[x, subject])[y].mean().reset_index()
#     obs_mean = df_obs.groupby(x)[y].mean()
#     obs_n = df_obs.groupby(x)[y].count()
#     x_loc = np.arange(len(df_obs[x].unique()))
#     if plot is None:
#         plot = 'violinplot' if has_pmt else 'barplot'
#     if plot == 'violinplot':
#         # Plot permutation distribution
#         dist_kws = dict(dict(color='gray', inner=None, linewidth=0), **(dist_kws if dist_kws is not None else {}))
#         sns.violinplot(x=x, y=y, data=df_pmt, order=x_order, **dist_kws)
#         # Plot originally observed data
#         scatter_kws = dict(dict(color=color, s=100, linewidths=1, edgecolors='k'), **(scatter_kws if scatter_kws is not None else {}))
#         plt.scatter(np.arange(len(x_order)), bounds.loc[x_order,'obs_mean'], **scatter_kws)
#     elif plot == 'lineplot':
#         line_kws = dict(dict(), **(line_kws if line_kws is not None else {}))
#         sns.lineplot(x=x, y=y, data=df_obs, ci=(1-alpha)*100, palette=color, **line_kws)
#         x_loc = df_obs[x].unique()
#     elif plot == 'barplot':
#         color = 'gray' if color is None else color
#         bar_kws = dict(dict(), **(bar_kws if bar_kws is not None else {}))
#         sns.barplot(x=x, y=y, data=df_obs, order=x_order, ci=(1-alpha)*100, color=color, **bar_kws)
#     elif plot == 'finalplot':
#         # Plot permutation distribution
#         dist_kws = dict(dict(color='gray', alpha=0.5, inner=None, linewidth=0), **(dist_kws if dist_kws is not None else {}))
#         sns.violinplot(x=x, y=y, data=df_pmt, order=x_order, **dist_kws)
#         # Plot bootstrap errorbars
#         color = 'k' if color is None else color
#         point_kws = dict(dict(linestyles='', scale=0.5, errwidth=2, capsize=0.1, facecolors='r'), **(point_kws if point_kws is not None else {}))
#         sns.pointplot(x=x, y=y, data=df_obs, order=x_order, ci=(1-alpha)*100, color=color, **point_kws)
#         # Plot originally observed data
#         scatter_kws = dict(dict(s=50, marker='o', linewidths=1, edgecolors=color, zorder=10), **(scatter_kws if scatter_kws is not None else {}))
#         plt.scatter(np.arange(len(x_order)), bounds.loc[x_order,'obs_mean'], **scatter_kws)
#     # Shift long ticklabels
#     if xtick_format is None:
#         xtick_format = ('normal' if plot == 'lineplot' else 'rotated')
#     if xtick_format == 'rotated':
#         plt.setp(plt.gca().get_xticklabels(), rotation=-30)
#         dx = 20/72; dy = 0/72 
#         offset = matplotlib.transforms.ScaledTranslation(dx, dy, fig.dpi_scale_trans)
#         for label in plt.gca().xaxis.get_majorticklabels():
#             label.set_transform(label.get_transform() + offset)
#     elif xtick_format == 'short':
#         plt.setp(plt.gca(), xticklabels=[label.get_text().split('_')[0] for label in plt.gca().get_xticklabels()])
#     elif xtick_format == 'normal':
#         pass
#     elif xtick_format == 'final':
#         plt.setp(plt.gca(), xticklabels=[label.get_text().split('_')[0] for label in plt.gca().get_xticklabels()])
#         plt.setp(plt.gca().get_xticklabels(), rotation=45, ha='right')
#         dx = 15/72; dy = 5/72 
#         offset = matplotlib.transforms.ScaledTranslation(dx, dy, fig.dpi_scale_trans)
#         for label in plt.gca().xaxis.get_majorticklabels():
#             label.set_transform(label.get_transform() + offset)
#     # Plot chance level
#     chance_kws = dict(dict(color='C3', ls='--', zorder=1), **(chance_kws if chance_kws is not None else {}))
#     plt.axhline(chance, **chance_kws)
#     # Plot reference line
#     if show_ref:
#         ref_kws = dict(dict(ref=0.55, color='gray', lw=0.5, ls='--'), **(ref_kws if ref_kws is not None else {}))
#         ref = ref_kws.pop('ref')
#         plt.axhline(ref, **ref_kws)
#     if star_alpha is None:
#         star_alpha = [0.3, 1]
#     if has_pmt:
#         # Plot multiple comparison correction band
#         if show_mcc:
#             plt.axhspan(bounds.loc['overall','lower'] if tail==2 else chance, bounds.loc['overall','upper'], color='r', alpha=0.1)
#         # Plot significant stars
#         if star_shift is None:
#             star_shift = bounds.ix[1:,'obs_std']/np.sqrt(bounds.ix[1:,'obs_n']) * 2.2 # Ignore first row
#         else:
#             star_shift = pd.Series(star_shift, index=bounds.index)
#         for k, x_label in enumerate(x_order):
#             if bounds.loc[x_label,'uncorrected']:
#                 plt.text(x_loc[k], bounds.loc[x_label,'obs_mean']+star_shift[x_label], '*', ha='center', alpha=star_alpha[0])
#             if bounds.loc[x_label,'corrected']:
#                 plt.text(x_loc[k], bounds.loc[x_label,'obs_mean']+star_shift[x_label], '*', ha='center', alpha=star_alpha[1])
#             if show_p:
#                 if mcc == 'none':
#                     plt.text(x_loc[k], 0.15, f"{bounds.loc[x_label,'p_uncorr']:.3f}", transform=myplot.transHDVA(), ha='center', alpha=star_alpha[0], fontsize='xx-small')
#                 if mcc == 'extreme':
#                     plt.text(x_loc[k], 0.15, f"{bounds.loc[x_label,'p_corr']:.3f}", transform=myplot.transHDVA(), ha='center', alpha=star_alpha[0], fontsize='xx-small')
#                 elif mcc == 'bonferroni':
#                     plt.text(x_loc[k], 0.15, f"{bounds.loc[x_label,'bonferroni']:.3f}", transform=myplot.transHDVA(), ha='center', alpha=star_alpha[1], fontsize='xx-small')
#     # Plot performance
#     if show_num:
#         for k, x_label in enumerate(x_order):
#             plt.text(x_loc[k], bounds.loc['overall','upper']*1.1-0.05 if has_pmt else 0.9, f"{obs_mean[x_label]:.3f}", 
#                 transform=plt.gca().transData if has_pmt else myplot.transHDVA(), ha='center', alpha=star_alpha[0], fontsize='xx-small')
#     # Plot obs_n
#     if show_n:
#         if len(set(obs_n[x_order].values)) == 1: # All equal
#             plt.text(0.95, 0.05, f"$n={obs_n[x_order[0]]}$", transform=plt.gca().transAxes, ha='right', fontsize='x-small')
#         else:
#             for k, x_label in enumerate(x_order):
#                 plt.text(x_loc[k], 0.05, f"$n={obs_n[x_label]}$" if k == 0 else f"${obs_n[x_label]}$", 
#                     transform=myplot.transHDVA(), ha='center', fontsize='x-small')


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import sys, os, re, glob, shlex, string
import subprocess, multiprocessing, ctypes, time, uuid
import json
import tables, warnings
from datetime import datetime
from itertools import chain
from collections import OrderedDict
from contextlib import contextmanager
from os import path
import numpy as np
from deepdish import io as dio
from . import six, afni


from .paraproc import format_duration
from .paraproc import cmd_for_exec, cmd_for_disp, run
from .paraproc import PooledCaller, SharedMemoryArray


package_dir = path.abspath(path.dirname(__file__))


def has_ants():
    pass


def has_N4():
    pass


def has_hcp_retino_docker():
    try:
        run(f"docker image list", goal_pattern='nben/neuropythy')
        return True
    except RuntimeError:
        print('>> Please check whether docker desktop is running.')
        return False


# Command line syntactic sugar
def expand_index_list(index_list, format=None):
    '''Expand a list of indices like: 1 3-5 7:10:2 8..10(2)
    '''
    flatten = []
    for x in index_list:
        if '-' in x: # 3-7 => 3,4,5,6,7 and 3-7-2 => 3,5,7
            range_ = list(map(int, x.split('-')))
            range_[1] += 1
            flatten.extend(range(*range_))
        elif ':' in x: # 3:7 => 3,4,5,6 and 3:7:2 => 3,5
            flatten.extend(range(*map(int, x.split(':'))))
        elif '..' in x: # 3..7 => 3,4,5,6,7 and '3..7(2)' => 3,5,7 (note that quotes are required in the shell because of the parentheses)
            range_ = list(map(int, re.split(r'\.\.|\(|\)', x)[:3])) # [:3] to get rid of the 4th, empty element
            range_[1] += 1
            flatten.extend(range(*range_))
        else: # 3
            flatten.append(int(x))
    return flatten if format is None else [format % x for x in flatten]




def select_and_replace_affix(glob_pattern, old_affix, new_affix):
    old_files = glob.glob(glob_pattern)
    pattern = re.compile(re.escape(old_affix))
    new_files = [re.sub(pattern, new_affix, f) for f in old_files]
    return old_files, new_files


def iterable(x): # Like the builtin `callable`
    try:
        iter(x)
        return not isinstance(x, six.string_types)
    except TypeError:
        return False


def contain_wildcard(fname):
    return any([w in fname for w in ['*', '?']])


class FilenameManager(object):
    @classmethod
    def from_glob(cls, fmt, **kwargs):
        self = cls(fmt, **kwargs)
        files = self.glob()
        self.parse(files)
        return self

    @classmethod
    def fmt2kws(cls, fmt, **kwargs):
        kws = {t[1]: kwargs.get(t[1], '*') for t in string.Formatter().parse(fmt) if t[1] is not None}
        return kws

    def __init__(self, fmt, **kwargs):
        self.fmt = fmt
        self.kws = self.fmt2kws(self.fmt, **kwargs)

    def glob(self, **kwargs):
        files = sorted(glob.glob(self.fmt.format(**dict(self.kws, **kwargs))))
        return files

    def format(self, fmt=None, keepdims=False, **kwargs):
        if fmt is None:
            fmt = self.fmt
        kws = self.fmt2kws(fmt, **dict(self.kws, **kwargs))
        kws = {k: (v if iterable(v) else [v]) for k, v in kws.items()}
        if not kws: # fmt contains no variable
            return fmt
        key2len = {k: len(v) for k, v in kws.items()}
        kws_lens = np.fromiter(key2len.values(), dtype=int)
        N = np.max(kws_lens)
        if not np.all(kws_lens[kws_lens>1] == N):
            raise ValueError('>> len() for different kwargs does not match.')
        files = []
        for n in range(N):
            kws_n = {k: (v[n] if key2len[k]>1 else v[0]) for k, v in kws.items()}
            files.append(fmt.format(**kws_n))
        if len(files) == 1 and not keepdims:
            return files[0]
        else:
            return files

    def _parse_one(self, fname):
        '''
        References
        ----------
        1. https://stackoverflow.com/questions/11844986/convert-or-unformat-a-string-to-variables-like-format-but-in-reverse-in-p
        '''
        fmt_escape = re.escape(self.fmt)
        fmt_regex = re.sub(r'\\{(.+?)\\}', r'(?P<\1>.+)', fmt_escape)
        fmt_regex_ = re.sub(r'\\{(.+?)\\}', r'(?P<\1>.+?)', fmt_escape)
        match = re.search(fmt_regex, fname)
        if match:
            kws = match.groupdict()
            match = re.search(fmt_regex_, fname)
            kws_ = match.groupdict()
            if kws != kws_:
                raise ValueError('>> Filename pattern is ambiguous.')
        else:
            raise ValueError('>> Filename pattern does not match provided files.')
        return kws

    def parse(self, files, multi_value='list'):
        '''
        files : list of filenames
        multi_value : {'wildcard', 'list'}
        '''
        all_kws = [self._parse_one(f) for f in files]
        kws = {}
        for key in all_kws[0]:
            values = [all_kws[k][key] for k in range(len(all_kws))]
            if np.all(np.array(values) == values[0]):
                kws[key] = all_kws[0][key]
            else:
                if multi_value == 'wildcard':
                    kws[key] = '*'
                elif multi_value == 'list':
                    kws[key] = values
        self.kws.update(kws)


def fname_with_ext(fname, ext):
    for k in reversed(range(len(ext)+1)):
        if fname.endswith(ext[:k]):
            return fname[:len(fname)-k] + ext


def exists(fname, force_redo=False):
    '''Lazy evaluation the body only if fname not already exists.'''
    if not path.exists(fname) or force_redo:
        return False
    else:
        print('>> Reuse existing "{0}"'.format(fname))
        return True


@contextmanager
def cd(d):
    old_dir = os.getcwd()
    os.chdir(d)
    yield
    os.chdir(old_dir)


class CacheManager(object):
    def __init__(self, persistent_file=None, ignore_init_run=True):
        self.persistent_file = 'cache_manager.json' if persistent_file is None else persistent_file
        self.ignore_init_run = ignore_init_run
        if self.persistent_file and path.exists(self.persistent_file):
            self.load_contexts()
        else:
            self.file_contexts = {}

    def load_contexts(self):
        if self.persistent_file:
            with open(self.persistent_file, 'r') as f:
                self.file_contexts = json.load(f)

    def save_contexts(self):
        if self.persistent_file:
            with open(self.persistent_file, 'w') as f:
                json.dump(self.file_contexts, f)

    def watch_files_updated(self, fname, watch_files):
        if watch_files is None:
            return False
        if isinstance(watch_files, six.string_types):
            watch_files = [watch_files]
        mtimes = self.file_contexts[fname]['watch_file_mtimes']
        flag = False
        for watch_file in watch_files:
            mtime = path.getmtime(watch_file)
            if watch_file not in mtimes:
                mtimes[watch_file] = mtime
                if not self.ignore_init_run:
                    print('>> File "{0}" is being watched'.format(watch_file))
                    flag = True
            elif mtime != mtimes[watch_file]:
                print('>> File "{0}" has been modified at {1}'.format(watch_file, 
                    datetime.fromtimestamp(mtime).strftime('%H:%M:%S %Y-%m-%d')))
                mtimes[watch_file] = mtime
                flag = True
        return flag

    def kwargs_updated(self, fname, kwargs):
        if not kwargs:
            return False
        kws = self.file_contexts[fname]['kwargs']
        flag = False
        for k, v in kwargs.items():
            if k not in kws:
                kws[k] = v
                if not self.ignore_init_run:
                    print('>> Variable "{0}" is being watched'.format(k))
                    flag = True
            elif v != kws[k]:
                print('>> Variable "{0}" has been modified from {1} -> {2}'.format(k, kws[k], v))
                kws[k] = v
                flag = True
        return flag

    def exists(self, fname, watch_files=None, force_redo=False, **kwargs):
        if fname not in self.file_contexts:
            self.file_contexts[fname] = {'watch_file_mtimes': {}, 'kwargs': {}}
        mtimes_updated = self.watch_files_updated(fname, watch_files)
        kws_updated = self.kwargs_updated(fname, kwargs)
        self.save_contexts()
        if not path.exists(fname) or mtimes_updated or kws_updated or force_redo:
            # Execute the code block
            return False
        else:
            print('>> Reuse existing "{0}"'.format(fname))
            return True


def temp_prefix(prefix='tmp_', n=4, suffix='.'):
    return prefix + uuid.uuid4().hex[:n] + suffix


def temp_folder(parent=None):
    folder = temp_prefix(suffix='')
    if parent is not None:
        folder = path.join(parent, folder)
    if not path.exists(folder):
        os.makedirs(folder)
    return folder


class ParallelCaller(object):
    def __init__(self):
        self.ps = []

    def check_call(self, cmd, **kwargs):
        '''
        Asynchronous check_call (launch and return immediately).
        Multiple commands can be stitched sequentially with ";" in linux/mac only if shell=True.
        '''
        cmd = cmd_for_exec(cmd, kwargs)
        if len(self.ps) == 0:
            self._start_time = time.time()
        print('>> job {0}: {1}'.format(len(self.ps), cmd_for_disp(cmd)))
        p = subprocess.Popen(cmd, **kwargs)
        self.ps.append(p)

    def wait(self):
        '''Wail for all parallel processes to finish (= wait for the slowest one).'''
        codes = [p.wait() for p in self.ps]
        duration = time.time() - self._start_time
        print('>> All {0} jobs done in {1}.'.format(len(self.ps), format_duration(duration)))
        self.ps = [] # Reset subprocess list
        return codes



def parallel_1D(cmd, in_file, prefix, n_jobs=1, combine_output=True, **kwargs):
    '''
    Parameters
    ----------
        cmd : str
            The command for parallel execution, must contain two placeholders
            {prefix} and {in_file}, which will be substituded with ''.format().
            As expected, other {} must be escaped as {{}}.
            For example,
                parallel_1D("3dTcat -prefix {prefix} -overwrite \
                    {in_file}'{{0:9}}'", 'xyz_list.1D', 'test', n_jobs=4)
    '''
    # Count the number of lines
    with open(in_file) as fi:
        n_lines = sum(1 for line in fi if line.strip() and not line.startswith('#'))
    # Split input 1D file
    files = [in_file+'-p{0:03d}'.format(j) for j in range(n_jobs)]
    job_size = (n_lines-1)//n_jobs + 1 # ceil
    with open(in_file) as fi:
        lines = (line for line in fi if line.strip() and not line.startswith('#'))
        for j in range(n_jobs):
            with open(files[j], 'w') as fj:
                for k in range(job_size):
                    try:
                        fj.write(next(lines))
                    except StopIteration:
                        pass
    # Parallel call
    prefixs = [prefix+'-p{0:03d}'.format(j) for j in range(n_jobs)]
    pc = ParallelCaller() # Execute n_jobs instances of cmd in parallel
    for j in range(n_jobs):
        pc.check_call(cmd.format(in_file=files[j], prefix=prefixs[j]), **kwargs)
    pc.wait() # Wait for all instances
    for j in range(n_jobs):
        os.remove(files[j])
    # Combine output 1D files
    if combine_output:
        outputs = glob.glob(prefix+'-*')
        match = re.match('{0}(.+)'.format(prefixs[0]), outputs[0])
        suffix = match.group(1)
        with open(prefix+suffix, 'w') as fo:
            for j in range(n_jobs):
                with open(outputs[j]) as fi:
                    for line in fi:
                        if line.strip() and not line.startswith('#'):
                            fo.write(line)
                os.remove(outputs[j])


def factorize(n):
    result = []
    for k in chain([2], range(3, n+1, 2)): # Test 2 and all odd integers between 3 and n
        while not n % k: # If n/k is an integer
            n = n // k
            result.append(k) # k is a prime factor
            if n == 1:
                return result
    return [1]


def parallel_3D(cmd, in_file, prefix, n_jobs=1, schema=None, fname_mapper=None, combine_output=True, **kwargs):
    '''
    Parameters
    ----------
    fname_mapper : dict or callable
    '''
    def n_factorize(factors, n):
        factors = sorted(factors, reverse=True) + [1]*(n-1)
        result = [1]*n
        for factor in factors:
            result[np.argmin(result)] *= factor
        return sorted(result, reverse=True)
    if fname_mapper is None:
        fname_mapper = lambda s: s
    elif isinstance(fname_mapper, dict):
        fname_map = fname_mapper
        def fname_mapper(s):
            for pattern, replace in fname_map.items():
                new_str, n = re.subn(pattern ,replace, s)
                if n > 0:
                    return new_str
            else:
                return s
    output_dir = path.dirname(prefix)
    # Determine split schema
    dims = afni.get_head_dims(in_file)[:3]
    if schema is None:
        factors = factorize(n_jobs)
        schema = np.r_[n_factorize(factors, 2), 1]
    dim_labels = ['RL', 'AP', 'IS']
    dim_order = np.argsort(dims)[::-1]
    split_schema = []
    for k in range(len(schema)):
        split_schema.append({'dim': dim_labels[dim_order[k]], 'splits': []})
        D = dims[dim_order[k]]
        delta = D // schema[k]
        for s in range(schema[k]):
            if s == schema[k] - 1:
                split_schema[k]['splits'].append([s*delta, 0])
            else:
                split_schema[k]['splits'].append([s*delta, D-(s+1)*delta])
    print('>> Split the volume into {0}x{1}x{2}={3} chunks...'.format(schema[0], schema[1], schema[2], np.prod(schema)))
    # Split input 3D file
    tmp_ = temp_prefix()
    s0, e0 = split_schema[0]['dim']
    s1, e1 = split_schema[1]['dim']
    s2, e2 = split_schema[2]['dim']
    split_params = {}
    pc = ParallelCaller()
    for m, (ns0, ne0) in enumerate(split_schema[0]['splits']):
        for n, (ns1, ne1) in enumerate(split_schema[1]['splits']):
            for l, (ns2, ne2) in enumerate(split_schema[2]['splits']):
                label = path.join(output_dir, '{0}{1:02d}_{2:02d}_{3:02d}'.format(tmp_, m, n, l))
                split_params[label] = (ns0, ne0, ns1, ne1, ns2, ne2, m, n, l)
                pc.check_call('''
                    3dZeropad -{s0} -{ns0} -{e0} -{ne0} -{s1} -{ns1} -{e1} -{ne1} -{s2} -{ns2} -{e2} -{ne2} \
                        -prefix {label} -overwrite {in_file}
                    '''.format(s0=s0, ns0=ns0, e0=e0, ne0=ne0, s1=s1, ns1=ns1, e1=e1, ne1=ne1, 
                        s2=s2, ns2=ns2, e2=e2, ne2=ne2, label=label, in_file=in_file))
    pc.wait()
    # Parallel call
    if not isinstance(cmd, six.string_types):
        cmd = ' '.join(cmd)
    for label in split_params:
        pc.check_call(cmd.format(in_file='{0}+orig'.format(label), prefix='{0}_out'.format(label)), **kwargs)
    pc.wait()
    # Combine output 3D files
    if combine_output:
        for label, (ns0, ne0, ns1, ne1, ns2, ne2, m, n, l) in split_params.items():
            fi = fname_mapper('{0}_out+orig'.format(label))
            pc.check_call('''
                3dZeropad -{s0} {ns0} -{e0} {ne0} -{s1} {ns1} -{e1} {ne1} -{s2} {ns2} -{e2} {ne2} \
                    -prefix {label}_pad -overwrite {fi}
                '''.format(s0=s0, ns0=ns0, e0=e0, ne0=ne0, s1=s1, ns1=ns1, e1=e1, ne1=ne1, 
                    s2=s2, ns2=ns2, e2=e2, ne2=ne2, label=label, fi=fi))
        pc.wait()
        glob_pattern = path.join(output_dir, '{0}*_pad+orig.HEAD'.format(tmp_))
        subprocess.check_call('''
            3dTstat -sum -prefix {prefix} -overwrite "{glob_pattern}"
            '''.format(prefix=prefix, glob_pattern=glob_pattern), shell=True)
    # Remove temp files
    for f in glob.glob(path.join(output_dir, tmp_+'*')):
        os.remove(f)


class Savable(object):
    def save(self, fname):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=tables.NaturalNameWarning)
            dio.save(fname, self.to_dict())

    @classmethod
    def load(cls, fname):
        return cls.from_dict(dio.load(fname))


class Savable2(object):
    def save(self, fname):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', category=tables.NaturalNameWarning)
            dio.save(fname, self.to_dict())

    def load(self, fname):
        self.from_dict(dio.load(fname))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import unittest
from mripy import utils
import time, ctypes
import numpy as np


class test_utils(unittest.TestCase):
    @unittest.skip('just for understanding whether and how it works')
    def test_SharedMemoryArray_memory(self):
        '''
        Test shared-memory versus copy-on-write
        '''
        def slow_operation(k, x):
            if isinstance(x, utils.SharedMemoryArray):
                x.arr[:,:100] += 1 # With copy-on-write, the difference becomes apparent when there is write access
                res = np.mean(x.arr) # Read access
            time.sleep(10)
            print(f'#{k}: mean = {res}')
        shape = (1000000,1000) # About 8 GB
        a = utils.SharedMemoryArray.zeros(shape) # This will be allocated in the Cached Files
        # The following two assignments are very different!
        # a.arr = np.random.rand(*shape)
        a.arr[:] = np.random.rand(*shape)
        pc = utils.PooledCaller()
        for k in range(pc.pool_size):
            pc.check_call(slow_operation, k, a)
        pc.wait()

    @unittest.skip('temporary')
    def test_SharedMemoryArray_array(self):
        '''
        Test the life saving container interface
        '''
        def slow_operation(k, x):
            while True:
                if x.acquire():
                    x[:,:100] += 1 # Write access
                    res = np.mean(x) # Read access
                    x.release()
                    break
            time.sleep(10)
            print(f'#{k}: mean = {res}')
        shape = (1000000,1000) # About 8 GB
        a = utils.SharedMemoryArray.zeros(shape)
        a[:] = np.random.rand(*shape)
        pc = utils.PooledCaller()
        for k in range(pc.pool_size):
            pc.check_call(slow_operation, k, a)
        pc.wait()

    def test_SharedMemoryArray_CoW(self):
        '''
        Test the speed of CoW if read-only 
        '''
        def slow_operation(k, x):
            res = np.mean(x) # Read access
            print(f'#{k}: mean = {res}')
        shape = (1000000,1000) # About 8 GB
        a = np.random.rand(*shape) # >> All 12 jobs done in 9.832 sec.
        # a = utils.SharedMemoryArray.zeros(shape) # >> All 12 jobs done in 8.784 sec.
        # a[:] = np.random.rand(*shape)
        pc = utils.PooledCaller()
        for k in range(pc.pool_size):
            pc.check_call(slow_operation, k, a)
        pc.wait()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import unittest
from numpy.testing import assert_allclose
try:
    from .context import data_dir # If mripy is importable: python -m mripy.tests.test_io
except ValueError: # Attempted relative import in non-package
    from context import data_dir # If not importable: cd mripy/tests; python -m test_io
from mripy import io

from os import path
import os, glob, subprocess
import numpy as np


class test_io(unittest.TestCase):
    def test_Mask(self):
        mask_file = path.join(data_dir, 'brain_mask', 'brain_mask+orig')
        data_file = path.join(data_dir, 'brain_mask', 'gre*.volreg+orig.HEAD')
        mask = io.Mask(mask_file)
        # Test dump
        x = mask.dump(data_file)
        self.assertEqual(x.shape, (564361,4))
        mask2 = io.MaskDumper(mask_file)
        y = mask2.dump(data_file)
        assert_allclose(x, y)
        # Test undump
        max_idx = np.argmax(x, axis=1) + 1
        mask.undump('test_undump', max_idx, method='nibabel')
        self.assertEqual(subprocess.check_output('3dinfo -orient test_undump+orig', shell=True), b'RSA\n')
        assert_allclose(mask.dump('test_undump+orig.HEAD'), max_idx)
        for f in glob.glob('test_undump+orig.*'):
            os.remove(f)
        # Test constrain
        smaller, sel0 = mask.near(5, 45, -17, 12, return_selector=True)
        sel = mask.infer_selector(smaller)
        smaller.undump('test_constrain', max_idx[sel])
        assert_allclose(smaller.dump('test_constrain+orig.HEAD'), max_idx[sel0])
        for f in glob.glob('test_constrain+orig.*'):
            os.remove(f)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import sys, os, subprocess
import re, glob, shlex, shutil, tempfile, warnings
import collections, itertools, copy
import random, string
from os import path
from datetime import datetime
import numpy as np
from scipy import ndimage
from . import six, utils, afni, math, paraproc
# For accessing NIFTI files
try:
    import nibabel
except ImportError:
    print('You may need to install "nibabel" to read/write NIFTI (*.nii) files.')
try:
    from lxml import etree
except ImportError:
    print('You may need to install "lxml" to read/write niml datasets (*.niml.dset).')

# Timestamp
def _timestamp(dt):
    '''
    Work-around for python 2.7 not having dt.timestamp() yet.
    http://stackoverflow.com/questions/11743019/convert-python-datetime-to-epoch-with-strftime
    '''
    return (dt - datetime(1970,1,1)).total_seconds()


def hms2dt(hms, date=None, timestamp=False):
    '''
    Convert time string in hms format to datetime object.

    `hms` is like "102907.165000". This format is used in dicom header.
    '''
    if date is None:
        date = '19700101'
    dt = datetime.strptime(date+hms, '%Y%m%d%H%M%S.%f')
    return _timestamp(dt) if timestamp else dt


def mmn2dt(mmn, date=None, timestamp=False):
    '''
    Convert time string in mmn format to datetime object.

    `mmn` is "msec since midnight", like "37747165". This format is used in
    physiological measurement log file.
    '''
    if date is None:
        date = '19700101'
    t = datetime.utcfromtimestamp(float(mmn)/1000)
    d = datetime.strptime(date, '%Y%m%d')
    dt = datetime.combine(d.date(), t.time())
    return _timestamp(dt) if timestamp else dt


# Physiological data
def _parse_physio_raw(fname):
    # print('Parsing "{0}"...'.format(fname))
    n_pre = {'ecg': 5, 'ext': 4, 'puls': 4, 'resp': 4}
    with open(fname, 'r') as fin:
        info = {}
        ch = path.splitext(fname)[1][1:]
        info['file'] = path.realpath(fname)
        info['channel'] = ch
        lines = fin.read().splitlines() # Without \n unlike fin.readlines()
        if len(lines) == 0:
            print('*+ WARNING: "{0}" seems be empty...'.format(fname), file=sys.stderr)
            return None
        k = 0
        # Data line(s)
        while lines[k][-4:] != '5003': # There can be more than one data lines
            k += 1
            if k >= len(lines): # The file does not contain 5003
                print('*+ WARNING: "{0}" might be broken...'.format(fname), file=sys.stderr)
                return None
        else:
            k += 1
            data_line = ''.join(lines[:k])
            info['messages'] = re.findall('\s5002\s(.+?)\s6002', data_line)
            data_line = re.sub('\s5002\s(.+?)\s6002', '', data_line) # Remove messages inserted between 5002/6002
            info['rawdata'] = np.int_(data_line.split()[n_pre[ch]:-1])
        # Timing lines
        items = ['LogStartMDHTime', 'LogStopMDHTime', 'LogStartMPCUTime', 'LogStopMPCUTime']
        for item in items:
            while True:
                match = re.match('({0}):\s+(\d+)'.format(item), lines[k])
                k += 1
                if match:
                    info[match.group(1)] = int(match.group(2))
                    break
        return info


def parse_physio_file(fname, date=None):
    '''
    Implementation notes
    --------------------
    1. The first 4 (ext, puls, resp) or 5 (ecg) values are parameters (of
       unknown meanings).
    2. There can be multiple data lines, within which extra parameters is
       inclosed between 5002 and 6002, especially for ecg.
    3. The footer is inclosed between 5003 and 6003, following physiological
       data (and that's why the final data value always appears to be 5003).
    4. The MDH values are timestamps derived from the clock in the scanner (so
       do DICOM images), while the MPCU values are timestamps derived from the
       clock within the PMU recording system [1]. Use MDH time to synchronize
       physiological and imaging time series.
    5. The trigger values (5000) are "inserted" into the data, and have to be
       stripped out from the time series [1]. This fact is double checked by
       looking at the smooth trend of the puls waveform.
    6. The sampling rate is slightly (and consistently) slower than specified
       in the manual and in [1].

    Notes about timing
    ------------------
    The scanner clock is slightly faster than the wall clock so that 2 sec in
    real time is recorded as ~2.008 sec in the scanner, affacting both dicom
    header and physiological footer, even though the actual TR is precisely 2 s
    (as measured by timing the s triggers with psychtoolbox) and the actual
    sampling rate of physiological data is precisely 50 Hz (as estimated by
    dividing the total number of samples by the corrected recording duration).

    References
    ----------
    [1] https://cfn.upenn.edu/aguirre/wiki/public:pulse-oximetry_during_fmri_scanning
    '''
    # fs = {'ecg': 400, 'ext': 200, 'puls': 50, 'resp': 50}
    fs = {'ecg': 398.4, 'ext': 199.20, 'puls': 49.80, 'resp': 49.80}
    trig_value = 5000
    tag_values = np.r_[trig_value, 6000]
    info = _parse_physio_raw(fname)
    if info is None:
        return None
    ch = info['channel']
    info['fs'] = fs[ch]
    info['start'] = mmn2dt(info['LogStartMDHTime'], date, timestamp=True)
    info['stop'] = mmn2dt(info['LogStopMDHTime'], date, timestamp=True)
    x = info['rawdata']
    if ch != 'ecg':
        # y = x.copy()
        # y = x[x!=trig_value] # Strip trigger value (5000)
        y = x[~np.in1d(x, tag_values)] # Strip all tag values (5000, 6000, ...)
        trig = np.zeros_like(x)
        trig[np.nonzero(x==trig_value)[0]-1] = 1
        trig = trig[x!=trig_value]
    else:
        y = x[:len(x)//2*2].reshape(-1,2)
        trig = np.zeros_like(y)
    info['data'] = y
    info['trig'] = trig
    info['t'] = info['start'] + np.arange(len(y)) / fs[ch]
    try:
        assert(max(y) < 4096) # Valid data range is [0, 4095]
    except AssertionError as err:
        print('\n** Invalid data value detected: {0}'.format(np.unique(y[y>4095])))
        raise err
    try:
        assert(np.abs(info['t'][-1]-info['stop'])<2/info['fs']) # Allow 1 sampleish error
    except AssertionError as err:
        print('\n** {0}: Last sample = {1}, stop = {2}, error = {3}'.format(
            info['channel'], info['t'][-1], info['stop'], info['t'][-1]-info['stop']))
        raise err
    return info


def parse_physio_files(fname, date=None, channels=None):
    '''
    '''
    if channels is None:
        channels = ['ecg', 'ext', 'puls', 'resp']
    stem = path.splitext(fname)[0]
    info = collections.OrderedDict()
    for ch in channels:
        info[ch] = parse_physio_file('.'.join((stem, ch)), date=date)
        if info[ch] is None:
            print('*+ WARNING: "{0}" info is missing. Skip "{1}"...'.format(ch, stem), file=sys.stderr)
            return None
    return info


def match_physio_with_series(physio_infos, series_infos, channel=None, method='cover'):
    if channel is None:
        channel = 'resp'
    physio_t = np.array([[p[channel]['start'], p[channel]['stop']] if p is not None else [0, 0] for p in physio_infos])
    physio = []
    series = []
    for k, s in enumerate(series_infos):
        if method == 'cover':
            p_idx = (physio_t[:,0] < s['start']) & (s['stop'] < physio_t[:,1])
            if np.any(p_idx):
                # If there is more than one (which should not be the case), use only the first one
                physio.append(physio_infos[np.nonzero(p_idx)[0][0]])
                series.append(s)
        elif method == 'overlap':
            p_idx = (physio_t[:,0] < s['stop']) & (s['start'] < physio_t[:,1]) # Thanks to Prof. Zhang Jun
            if np.any(p_idx):
                # If there is more than one (which should not be the case), use the one with largest overlap
                overlap = np.maximum(physio_t[p_idx,0], s['start']) - np.minimum(physio_t[p_idx,1], s['stop'])
                idx = np.nonzero(p_idx)[0][np.argmax(overlap)]
                physio.append(physio_infos[idx])
                series.append(s)
    return physio, series


def _print_physio_timing(pinfo, sinfo, channel, index=None):
    prefix = channel if index is None else '#{0} ({1})'.format(index, channel)
    print('{0}: pre={1:.3f}, scan={2:.3f}, post={3:.3f}, total={4:.3f}'.format(
        prefix, sinfo['start']-pinfo['start'], sinfo['stop']-sinfo['start'],
        pinfo['stop']-sinfo['stop'], pinfo['stop']-pinfo['start']))


def extract_physio(physio_file, dicom_file, TR=None, dummy=0, channels=['resp', 'puls'], verbose=1):
    sinfo = parse_series_info(dicom_file) if isinstance(dicom_file, six.string_types) else dicom_file
    pinfo = parse_physio_files(physio_file, date=sinfo['date']) if isinstance(physio_file, six.string_types) else physio_file
    res = []
    for ch in channels:
        info = pinfo[ch]
        t = info['t']
        valid = (t >= sinfo['start']+dummy*TR) & (t < sinfo['stop']) # Assume timestamp indicates the start of the volume
        res.append(info['data'][valid])
        if verbose:
            _print_physio_timing(info, sinfo, ch)
    return res


# DICOM
def parse_dicom_header(fname, fields=None):
    '''
    Execute afni command `dicom_hdr` to readout most useful info from dicom header.

    Parameters
    ----------
    fname : str
    fields : {field: (matcher, extracter(match))}
        You can require additional fields in dicom header to be parsed.
        - field : e.g., 'ImageTime'
        - matcher : e.g. r'ID Image Time//(\S+)'
        - extracter : e.g., lambda match: io.hms2dt(match.group(1), date='20170706', timestamp=True)
    '''
    # print(fname)
    header = collections.OrderedDict()
    lines = subprocess.check_output(['dicom_hdr', fname]).decode('utf-8').split('\n')
    k = 0
    try:
        while True:
            match = re.search(r'ID Acquisition Date//(\S+)', lines[k])
            k += 1
            if match:
                header['AcquisitionDate'] = match.group(1)
                break
        while True:
            match = re.search(r'ID Acquisition Time//(\S+)', lines[k])
            k += 1
            if match:
                header['AcquisitionTime'] = match.group(1) # This marks the start of a volume
                header['timestamp'] = hms2dt(header['AcquisitionTime'], date=header['AcquisitionDate'], timestamp=True)
                break
        while True:
            match = re.search(r'ACQ Scanning Sequence//(.+)', lines[k])
            k += 1
            if match:
                header['sequence_type'] = match.group(1).strip()
                break
        while True:
            match = re.search(r'ACQ Sequence Variant//(.+)', lines[k])
            k += 1
            if match:
                header['sequence_type'] = ' '.join((match.group(1).strip(), header['sequence_type']))
                break
        while True:
            match = re.search(r'ACQ MR Acquisition Type //(.+)', lines[k])
            k += 1
            if match:
                header['sequence_type'] = ' '.join((match.group(1).strip(), header['sequence_type']))
                break
        while True:
            match = re.search(r'ACQ Slice Thickness//(\S+)', lines[k])
            k += 1
            if match:
                header['resolution'] = [float(match.group(1))]
                break
        while True:
            match = re.search(r'ACQ Repetition Time//(\S+)', lines[k])
            k += 1
            if match:
                header['RepetitionTime'] = float(match.group(1)) # ms
                break
        while True:
            match = re.search(r'ACQ Echo Time//(\S+)', lines[k])
            k += 1
            if match:
                header['TE'] = float(match.group(1)) # ms
                break
        while True:
            match = re.search(r'ACQ Imaging Frequency//(\S+)', lines[k])
            k += 1
            if match:
                header['Larmor'] = float(match.group(1)) # MHz
                break
        while True:
            match = re.search(r'ACQ Echo Number//(\S+)', lines[k])
            k += 1
            if match:
                header['EchoNumber'] = int(match.group(1)) # For multi-echo images
                break
        while True:
            match = re.search(r'ACQ Magnetic Field Strength//(\S+)', lines[k])
            k += 1
            if match:
                header['B0'] = float(match.group(1)) # Tesla
                break
        while True:
            match = re.search(r'ACQ Pixel Bandwidth//(\S+)', lines[k])
            k += 1
            if match:
                header['BW'] = float(match.group(1)) # Hz/pixel
                break
        while True:
            match = re.search(r'ACQ Protocol Name//(.+)', lines[k])
            k += 1
            if match:
                header['ProtocolName'] = match.group(1).strip()
                break
        while True:
            match = re.search(r'ACQ Flip Angle//(\S+)', lines[k])
            k += 1
            if match:
                header['FlipAngle'] = float(match.group(1))
                break
        while True:
            match = re.search(r'ACQ SAR//(\S+)', lines[k])
            k += 1
            if match:
                header['SAR'] = float(match.group(1))
                break
        while True: # This field is optional
            match = re.search(r'0019 100a.+//\s*(\d+)', lines[k])
            k += 1
            if match:
                header['n_slices'] = int(match.group(1))
                break
            if lines[k].startswith('0020'):
                break
        while True:
            match = re.search(r'REL Study ID//(\d+)', lines[k])
            k += 1
            if match:
                header['StudyID'] = int(match.group(1)) # Study index
                break
        while True:
            match = re.search(r'REL Series Number//(\d+)', lines[k])
            k += 1
            if match:
                header['SeriesNumber'] = int(match.group(1)) # Series index
                break
        while True:
            match = re.search(r'REL Acquisition Number//(\d+)', lines[k])
            k += 1
            if match:
                header['AcquisitionNumber'] = int(match.group(1)) # Volume index
                break
        while True:
            match = re.search(r'REL Instance Number//(\d+)', lines[k])
            k += 1
            if match:
                header['InstanceNumber'] = int(match.group(1)) # File index (whether it is one volume or one slice per file)
                break
        while True:
            match = re.search(r'IMG Pixel Spacing//(\S+)', lines[k])
            k += 1
            if match:
                header['resolution'] = list(map(float, match.group(1).split('\\'))) + header['resolution']
                break
        while True: # This field is optional
            match = re.search(r'0051 1011.+//(\S+)', lines[k])
            k += 1
            if match:
                header['iPAT'] = match.group(1)
                break
            if lines[k].startswith('Group'):
                break
    except IndexError as error:
        print('** Failed to process "{0}"'.format(fname))
        raise error
    if fields is not None:
        for line in lines:
            for field, (matcher, extracter) in fields.items():
                match = re.search(matcher, line)
                if match:
                    header[field] = extracter(match)
                    break
    header['gamma'] = 2*np.pi*header['Larmor']/header['B0']
    return header


SERIES_PATTERN = r'.+?\.(\d{4})\.' # Capture series number
MULTI_SERIES_PATTERN = r'.+?\.(\d{4})\.(\d{4}).+(\d{8,})' # Capture series number, slice number, uid
MULTI_SERIES_PATTERN2 = r'.+?\.(\d{4})\.(\d{4}).+(\d{5,}\.\d{8,})' # Capture series number, slice number, uid (5-6.8-9)
# MULTI_SERIES_PATTERN3 = r'.+?\.(\d{4})\.(\d{4}).+(\d{5,})\.\d{5,}' # Capture series number, slice number, uid (5-6).5-9
MULTI_SERIES_PATTERN3 = r'.+?\.(\d{4})\.(\d{4}).+(\d+)\.\d+' # Capture series number, slice number, uid

def _sort_multi_series(files):
    '''
    Sort multiple series sharing the same series number into different studies.
    '''
    series = []
    timestamps = []
    infos = []
    for f in files:
        match = re.search(MULTI_SERIES_PATTERN3, f)
        # infos.append((f, int(match.group(2)), int(match.group(3))))
        infos.append((f, int(match.group(2)), float(match.group(3))))
    prev_slice = sys.maxsize
    for f, curr_slice, timestamp in sorted(infos, key=lambda x: x[-1]):
        if curr_slice <= prev_slice and (prev_slice == sys.maxsize or curr_slice in slices):
            # We meet a new sequence (including the first one).
            # Note that slices within a study are unique but may not be strictly ordered.
            # The first clause is a shortcut, and the second one is the real condition.
            series.append([])
            slices = set()
            timestamps.append(timestamp)
        series[-1].append(f)
        slices.add(curr_slice)
        prev_slice = curr_slice
    return series, timestamps


def sort_dicom_series(folder, series_pattern=SERIES_PATTERN):
    '''
    Parameters
    ----------
    folder : string
        Path to the folder containing all the *.IMA files.

    Returns
    -------
    studies : list of dicts
        [{'0001': [file0, file1, ...], '0002': [files], ...}, {study1}, ...]
    '''
    # Sort files into series
    files = sorted(glob.glob(path.join(folder, '*.IMA')))
    series = collections.OrderedDict()
    for f in files:
        filename = path.basename(f)
        match = re.search(series_pattern, filename)
        sn = match.group(1)
        if sn not in series:
            series[sn] = []
        series[sn].append(filename) # Changed 2019-10-25: series[sn].append(f)
    # Separate potentially multiple series sharing the same series number into different studies
    studies = None
    for s_idx, (sn, files) in enumerate(series.items()):
        subsets, timestamps = _sort_multi_series(files)
        if s_idx == 0:
            n_folders = len(subsets)
            # Note that if the first series is single, all series must be single.
            if n_folders == 1:
                studies = [series]
                break
            else:
                studies = [collections.OrderedDict() for k in range(n_folders)]
            for k, subset in enumerate(subsets):
                studies[k][sn] = subset
            start_times = timestamps
        else:
            # Handle the case when a later study has more series than earlier studies
            for k, subset in enumerate(subsets):
                kk = n_folders - 1
                while start_times[kk] > timestamps[k]:
                    kk -= 1
                studies[kk][sn] = subset
    return studies


def filter_dicom_files(files, series_numbers=None, instance_numbers=None, series_pattern=MULTI_SERIES_PATTERN3):
    if isinstance(files, six.string_types) and path.isdir(files):
        files = glob.glob(path.join(files, '*.IMA'))
    if not isinstance(series_numbers, collections.Iterable):
        series_numbers = [series_numbers]
    if not isinstance(instance_numbers, collections.Iterable):
        instance_numbers = [instance_numbers]
    files = np.array(sorted(files))
    if len(files) == 0:
        return []
    infos = []
    for fname in files:
        filepath, filename = path.split(fname)
        match = re.match(series_pattern, filename)
        infos.append(list(map(int, match.groups()))) # series number, instance number, uid
    infos = np.array(infos)
    filtered = []
    if not series_numbers: # [] or None
        series_numbers = np.unique(infos[:,0])
    for series in series_numbers:
        if not instance_numbers:
            instance_numbers = np.unique(infos[infos[:,0]==series,1])
        for instance in instance_numbers:
            filtered.extend(files[(infos[:,0]==series)&(infos[:,1]==instance)])
    return filtered


def parse_slice_order(dicom_files):
    t = None
    if len(dicom_files) > 1:
        temp_dir = 'temp_pares_slice_order'
        os.makedirs(temp_dir)
        for k, f in enumerate(dicom_files[:2]):
            shutil.copyfile(f, path.join(temp_dir, '{0}.IMA'.format(k)))
        old_path = os.getcwd()
        try:
            os.chdir(temp_dir)
            afni.check_output('''Dimon -infile_pattern '*.IMA'
                -gert_create_dataset -gert_to3d_prefix temp -gert_quit_on_err''')
            res = afni.check_output(['3dAttribute', 'TAXIS_OFFSETS', 'temp+orig'])[-2]
            t = np.array(list(map(float, res.split())))
        finally:
            os.chdir(old_path)
            shutil.rmtree(temp_dir)
    if t is None:
        order = None
    elif np.all(np.diff(t) > 0):
        order = 'ascending'
    elif np.all(np.diff(t) < 0):
        order = 'descending'
    else:
        order = 'interleaved'
    return order, t


def parse_series_info(fname, timestamp=False, shift_time=None, series_pattern=SERIES_PATTERN, fields=None, parser=None):
    '''
    Potential bug: `dicom.parse_dicom_header` doesn't support `fields` as kwargs 
    '''
    if isinstance(fname, six.string_types): # A single file or a folder
        if path.isdir(fname):
            # Assume there is only one series in the folder, so that we only need to consider the first file.
            fname = sorted(glob.glob(path.join(fname, '*.IMA')))[0]
        # Select series by series number (this may fail if there is multi-series in the folder)
        filepath, filename = path.split(fname)
        match = re.match(series_pattern, filename)
        files = sorted(glob.glob(path.join(filepath, '{0}*.IMA'.format(match.group(0)))))
        findex = None
    else: # A list of files (e.g., as provided by sort_dicom_series)
        files = fname
        findex = 0
    if parser is None:
        parser = parse_dicom_header
    info = collections.OrderedDict()
    if timestamp:
        parse_list = range(len(files))
    else:
        parse_list = [0, -1]
    headers = [parser(files[k], fields=fields) for k in parse_list]
    if headers[0]['StudyID'] != headers[-1]['StudyID']:
        # There are more than one series (from different studies) sharing the same series number
        if parse_list == [0, -1]:
            headers = [headers[0]] + [parser(f) for f in files[1:-1]] + [headers[-1]]
        if findex is None:
            findex = files.index(fname)
        selected = [k for k, header in enumerate(headers) if header['StudyID']==headers[findex]['StudyID']]
        files = [files[k] for k in selected]
        headers = [headers[k] for k in selected]
    info.update(headers[0])
    info['date'] = info['AcquisitionDate']
    info['first'] = headers[0]['timestamp']
    info['last'] = headers[-1]['timestamp']
    info['n_volumes'] = headers[-1]['AcquisitionNumber'] - headers[0]['AcquisitionNumber'] + 1
    info['TR'] = (info['last']-info['first'])/(info['n_volumes']-1) if info['n_volumes'] > 1 else None
    if shift_time == 'CMRR':
        shift_time = 0
        if info['TR'] is not None and 'n_slices' in info and np.mod(info['n_slices'], 2)==0:
            slice_order = parse_slice_order(files)[0]
            if slice_order == 'interleaved':
                shift_time = -info['TR']/2
    elif shift_time is None:
        shift_time = 0
    info['first'] += shift_time
    info['last'] += shift_time
    info['start'] = info['first']
    info['stop'] = (info['last'] + info['TR']) if info['TR'] is not None else info['last']
    if timestamp:
        info['t'] = np.array([header['timestamp'] for header in headers]) + shift_time
    info['files'] = [path.realpath(f) for f in files]
    info['headers'] = headers
    return info


def convert_dicom(dicom_dir, out_file=None, dicom_ext=None, interactive=False):
    if dicom_ext is None:
        dicom_ext = '.IMA'
    if out_file is None:
        out_file = './*.nii'
    out_dir, prefix, ext = afni.split_out_file(out_file, split_path=True)
    out_dir = path.realpath(path.expanduser(out_dir)) # It is special here because we'll change dir later
    if not path.exists(out_dir):
        os.makedirs(out_dir)
    if prefix == '*': # Take the dicom folder name by default
        prefix = path.split(dicom_dir)[1]
    old_path = os.getcwd()
    try:
        os.chdir(dicom_dir)
        with open('uniq_image_list.txt', 'w') as fo:
            subprocess.check_call(['uniq_images'] + glob.glob('*'+dicom_ext), stdout=fo) # Prevent shell injection by not using shell=True with user defined string
        interactive_cmd = '' if interactive else '-gert_quit_on_err'
        utils.run("Dimon -infile_list uniq_image_list.txt \
            -gert_create_dataset -gert_outdir '{0}' -gert_to3d_prefix '{1}' -overwrite \
            -dicom_org -use_obl_origin -save_details Dimon.details {2}".format(out_dir, prefix+ext, interactive_cmd))
    finally:
        os.chdir(old_path)


def convert_dicoms(dicom_dirs, out_dir=None, prefix=None, out_type='.nii', dicom_ext='.IMA', **kwargs):
    '''
    Parameters
    ----------
    dicom_dirs : list or str
        1. A list of folders containing *.IMA files
        2. It can also be a glob pattern that describes a list of folders, e.g., "raw_fmri/func??"
        3. Finally, it can be a root folder (e.g., "raw_fmri") containing multiple sub-folders of *.IMA files, 
           raw_fmri/anat, raw_fmri/func01, raw_fmri/func02, etc.
    out_dir : str
        Output directory for converted datasets, default is current directory.
        The output would look like:
            out_dir/anat.nii, out_dir/func01.nii, out_dir/func02.nii, etc.
    '''
    original_dicom_dirs = dicom_dirs
    if isinstance(dicom_dirs, six.string_types):
        if utils.contain_wildcard(dicom_dirs):
            dicom_dirs = glob.glob(dicom_dirs)
        else:
            dicom_dirs = glob.glob(path.join(dicom_dirs, '*'))
    if len(dicom_dirs) == 0: # Sanity check after Yong's true story
        warnings.warn(f"\n>> Cannot find any dicom file to convert. Is the following path correct?\n{original_dicom_dirs}")
    if out_dir is None:
        out_dir = '.'
    idx = 0
    for f in dicom_dirs:
        if path.isdir(f) and len(glob.glob(path.join(f, '*'+dicom_ext))) > 0:
            idx += 1
            convert_dicom(f, path.join(out_dir, '*'+out_type if prefix is None else '{0}{1:02d}{2}'.format(prefix, idx, out_type)), dicom_ext=dicom_ext, **kwargs)


# ========== Generic read/write ==========
def read_vol(fname, return_img=False):
    img = nibabel.load(fname)
    vol = np.asanyarray(img.dataobj) # Equivalent to the deprecated "vol = img.get_data()" that gives minimum possible data size in memory
    # vol = img.get_fdata() # get_data() is deprecated in favor of get_fdata(), which has a more predictable return type
    return (vol, img) if return_img else vol


def write_vol(fname, vol, base_img=None):
    if fname.endswith('.nii'):
        write_nii(fname, vol, base_img)
    else:
        write_afni(fname, vol, base_img)


def read_surf_mesh(fname, return_img=False, **kwargs):
    if fname.endswith('.asc'):
        verts, faces = read_asc(fname, **kwargs)
        img = None
    elif fname.endswith('.gii'):
        verts, faces, img = read_gii(fname, return_img=True)
    return (verts, faces, img) if return_img else (verts, faces)


def write_surf_mesh(fname, verts, faces, **kwargs):
    if fname.endswith('.asc'):
        write_asc(fname, verts, faces, **kwargs)
    elif fname.endswith('.gii'):
        write_gii(fname, verts, faces, **kwargs)


def read_surf_data(fname):
    if fname.endswith('.niml.dset'):
        nodes, values = read_niml_bin_nodes(fname)
    return nodes, values


def write_surf_data(fname, nodes, values):
    if fname.endswith('.niml.dset'):
        write_niml_bin_nodes(fname, nodes, values)


def read_surf_info(fname):
    info = {}
    if fname.endswith('.asc'):
        with open(fname) as fi:
            for line in fi:
                if not line.startswith('#'):
                    info['n_verts'], info['n_faces'] = np.int_(line.split())
                    break
        info['hemi'] = afni.get_hemi(fname)
        info['ext'] = '.asc'
    elif fname.endswith('.gii'):
        img = nibabel.load(fname)
        info['n_verts'] = img.get_arrays_from_intent(nibabel.nifti1.intent_codes['NIFTI_INTENT_POINTSET'])[0].dims[0]
        info['n_faces'] = img.get_arrays_from_intent(nibabel.nifti1.intent_codes['NIFTI_INTENT_TRIANGLE'])[0].dims[0]
        info['hemi'] = afni.get_hemi(fname)
        info['ext'] = '.gii'
    return info


def read_txt(fname, dtype=float, comment='#', delimiter=None, skiprows=0, nrows=None, return_comments=False):
    '''Read numerical array from text file, much faster than np.loadtxt()'''
    with open(fname, 'r') as fin:
        lines = fin.readlines()
    if return_comments:
        comments = [line for line in lines[skiprows:] if line.strip() and line.startswith(comment)]
    lines = [line for line in lines[skiprows:(nrows if not nrows else skiprows+nrows)] if line.strip() and not line.startswith(comment)]
    n_cols = len(lines[0].split(delimiter))
    x = np.fromiter(itertools.chain.from_iterable(
        map(lambda line: line.split(delimiter), lines)), dtype=dtype).reshape(-1,n_cols)
    if return_comments:
        return x, comments
    else:
        return x


def read_stim(fname):
    with open(fname) as fi:
        return [np.array([]) if line[0] == '*' else np.float_(line.split()) for line in fi if line.strip()]


# ========== NIFTI ==========
def read_nii(fname, return_img=False):
    if fname[-4:] != '.nii' and fname[-7:] != '.nii.gz':
        fname = fname + '.nii'
    img = nibabel.load(fname)
    vol = img.get_data()
    return (vol, img) if return_img else vol


def write_nii(fname, vol, base_img=None, space=None, dim=None):
    if fname[-4:] != '.nii' and fname[-7:] != '.nii.gz':
        fname = fname + '.nii'
    if base_img is None:
        affine = nibabel.affines.from_matvec(np.eye(3), np.zeros(3))
    elif isinstance(base_img, six.string_types):
        affine = nibabel.load(base_img).affine
    else:
        affine = base_img.affine
    if dim is not None:
        # img.header['dim'] = dim # This header field will be overwritten by vol.shape
        vol = vol.reshape(dim[1:1+dim[0]])
    img = nibabel.Nifti1Image(vol, affine)
    # https://afni.nimh.nih.gov/afni/community/board/read.php?1,149338,149340#msg-149340
    # 0 (unknown) sform not defined
    # 1 (scanner) RAS+ in scanner coordinates
    # 2 (aligned) RAS+ aligned to some other scan
    # 3 (talairach) RAS+ in Talairach atlas space
    # 4 (mni) RAS+ in MNI atlas space
    if space is None:
        space = 1 if base_img is None else base_img.header['sform_code']
    img.header['sform_code'] = space
    nibabel.save(img, fname)


SPACE_CODE = {
    'unknown': 0,
    'scanner': 1, 'orig': 1, 'ORIG': 1,
    'aligned': 2,
    'talairach': 3, 'tlrc': 3, 'TLRC': 3,
    'mni': 4, 'MNI': 4,
}


def get_space(in_file):
    space = int(nibabel.load(in_file).header['sform_code'])
    return {0: 'unknown', 1: 'orig', 2: 'aligned', 3: 'tlrc', 4: 'mni'}[space]


def change_space(in_file, out_file=None, space=None, method='nibabel'):
    '''
    >>> change_space('MNI152_2009_template.nii.gz', 'template.nii', space='ORIG')
    >>> change_space('test+tlrc.HEAD') # -> test.nii as ORIG
    '''
    if isinstance(space, str):
        space = SPACE_CODE[space]
    if method == 'nibabel':
        if out_file is None:
            prefix, ext = afni.split_out_file(in_file)
            out_file = f"{prefix}.nii"
        vol, img = read_vol(in_file, return_img=True)
        # write_nii(out_file, vol, base_img=img, space=space)
        # To work-around a strang bug in nibabel when overwriting an existing volume in linux,
        # an explicit copy of the memmap has to be made.
        write_nii(out_file, vol.copy(), base_img=img, space=space)
    elif method == 'afni':
        if space is None:
            space = 1
        # afni.set_nifti_field(in_file, 'sform_code', space, out_file=out_file)
        raise NotImplementedError()
        # Error message:
        # ** ERROR: EDIT_dset_items[244]: illegal new xyzdel
        # ** ERROR: EDIT_dset_items[244]: illegal new xyzorient
        # Before (with `nifti_tool -disp_hdr -field srow_x ...`):
        # srow_x               280      4    -0.8 -0.0 -0.0 62.294399
        # Aftre:
        # srow_x               280      4    0.0 0.0 0.0 0.0
        # With nibabel:
        # srow_x               280      4    -0.7 0.000183 -0.000183 110.674217


def get_dim_order(in_file):
    dim = nibabel.load(in_file).header['dim']
    if dim[0] > 4:
        return 'bucket'
    else:
        return 'timeseries'


def change_dim_order(in_file, out_file=None, dim_order=None, method='afni'):
    '''
    dim_order : 1D array with 8 numbers
        e.g., np.array([  5, 300, 300, 124,   1,   2,   1,   1], dtype=np.int16) # for stats 
        or, np.array([  4, 150, 150,  62, 158,   1,   1,   1], dtype=np.int16) # for epi
    method : str, 'afni' | 'nibabel'
    '''
    if dim_order is None:
        dim_order = 'timeseries'
    def get_new_dim(dim, dim_order):
        if isinstance(dim_order, str):
            if dim_order in ['timeseries']:
                new_dim = np.r_[4, dim[1:4], max(dim[4:]), 1, 1, 1]
            elif dim_order in ['stats', 'bucket']:
                new_dim = np.r_[5, dim[1:4], 1, max(dim[4:]), 1, 1]
            else:
                raise ValueError('** Only support "timeseries" and "bucket" dim_order.')
        else:
            new_dim = dim_order
        return new_dim.astype(np.int16)
    if method == 'nibabel':
        if out_file is None:
            prefix, ext = afni.split_out_file(in_file)
            out_file = f"{prefix}.nii"
        vol, img = read_vol(in_file, return_img=True)
        dim = img.header['dim']
        write_nii(out_file, vol, base_img=img, dim=get_new_dim(dim, dim_order))
    elif method == 'afni':
        dim = afni.get_nifti_field(in_file, 'dim', 'int')
        afni.set_nifti_field(in_file, 'dim', get_new_dim(dim, dim_order), out_file=out_file)


# ========== AFNI HEAD/BRIK ==========
def read_afni(fname, remove_nii=True, return_img=False):
    try:
        if fname[-5:] in ['.HEAD', '.BRIK']:
            pass
        elif fname[-1] == '.':
            fname = fname + 'HEAD'
        else:
            fname = fname + '.HEAD'
        img = nibabel.load(fname) # Start from nibabel 2.3.0 (with brikhead.py)
        # vol = img.get_data().squeeze()
        vol = img.get_data()
        return (vol, img) if return_img else vol
    except nibabel.filebasedimages.ImageFileError:
        print('*+ WARNING: Fail to open "{0}" with nibabel, fallback to 3dAFNItoNIFTI'.format(fname)) 
        match = re.match('(.+)\+', fname)
        nii_fname = match.group(1) + '.nii'
        subprocess.check_call(['3dAFNItoNIFTI', '-prefix', nii_fname, fname])
        res = read_nii(nii_fname, return_img)
        if remove_nii:
            os.remove(nii_fname)
        return res


def write_afni(prefix, vol, base_img=None):
    nii_fname = prefix + '.nii'
    write_nii(nii_fname, vol, base_img)
    subprocess.check_call(['3dcopy', nii_fname, prefix+'+orig', '-overwrite'])
    os.remove(nii_fname)




# ========== AFNI ASC ==========
def read_asc(fname, dtype=None):
    '''Read FreeSurfer/SUMA surface (vertices and faces) in *.asc format.'''
    if dtype is None:
        dtype = float
    with open(fname, 'r') as fin:
        lines = fin.readlines()
    n_verts, n_faces = np.int_(lines[1].split())
    # verts = np.vstack(map(lambda line: np.float_(line.split()), lines[2:2+n_verts])) # As slow as np.loadtxt()
    # verts = np.float_(''.join(lines[2:2+n_verts]).split()).reshape(-1,4) # Much faster
    verts = np.fromiter(itertools.chain.from_iterable(
        map(lambda line: line.split()[:3], lines[2:2+n_verts])), dtype=dtype).reshape(-1,3)
    faces = np.fromiter(itertools.chain.from_iterable(
        map(lambda line: line.split()[:3], lines[2+n_verts:2+n_verts+n_faces])), dtype=int).reshape(-1,3)
    return verts, faces


def read_patch_asc(fname, dtype=None, index_type='multimap'):
    '''
    Read FreeSurfer/SUMA patch (noncontiguous vertices and faces) in *.asc format.
    
    index_type : str
        - "raw" or "array"
        - "map" or "dict"
        - "multimap" or "func"
    '''
    if dtype is None:
        dtype = float
    with open(fname, 'r') as fin:
        lines = fin.readlines()
    n_verts, n_faces = np.int_(lines[1].split())
    # verts = np.vstack(map(lambda line: np.float_(line.split()), lines[2:2+n_verts])) # As slow as np.loadtxt()
    # verts = np.float_(''.join(lines[2:2+n_verts]).split()).reshape(-1,4) # Much faster
    verts = np.fromiter(itertools.chain.from_iterable(
        map(lambda line: line.split()[:3], lines[2+1:2+n_verts*2:2])), dtype=dtype).reshape(-1,3)
    faces = np.fromiter(itertools.chain.from_iterable(
        map(lambda line: line.split()[:3], lines[2+1+n_verts*2:2+n_verts*2+n_faces*2:2])), dtype=int).reshape(-1,3)
    vidx = np.fromiter(itertools.chain.from_iterable(
        map(lambda line: line.split('=')[-1:], lines[2:2+n_verts*2:2])), dtype=int)
    fidx = np.fromiter(itertools.chain.from_iterable(
        map(lambda line: line.split('=')[-1:], lines[2+n_verts*2:2+n_verts*2+n_faces*2:2])), dtype=int)
    vmap = {vidx[k]: k for k in range(n_verts)}
    fmap = {fidx[k]: k for k in range(n_faces)}
    if index_type in ['raw', 'array']:
        pass
    if index_type in ['map', 'dict']:
        vidx = vmap
        fidx = fmap
    elif index_type in ['multimap', 'func']:
        vidx = lambda K: vmap[K] if np.isscalar(K) else [vmap[k] for k in K]
        fidx = lambda K: fmap[K] if np.isscalar(K) else [fmap[k] for k in K]
    return verts, faces, vidx, fidx


def write_asc(fname, verts, faces):
    with open(fname, 'wb') as fout: # Binary mode is more compatible with older Python...
        fout.write('#!ascii version of surface mesh saved by mripy\n'.encode('ascii'))
        np.savetxt(fout, [[len(verts), len(faces)]], fmt='%d')
        np.savetxt(fout, np.c_[verts, np.zeros(len(verts))], fmt=['%.6f', '%.6f', '%.6f', '%d'])
        np.savetxt(fout, np.c_[faces, np.zeros(len(faces))], fmt='%d')    


# ========== GIFTI ==========
def read_gii(fname, return_img=False):
    img = nibabel.load(fname)
    # verts, faces = img.darrays[0].data, img.darrays[1].data
    verts = img.get_arrays_from_intent(nibabel.nifti1.intent_codes['NIFTI_INTENT_POINTSET'])[0].data
    faces = img.get_arrays_from_intent(nibabel.nifti1.intent_codes['NIFTI_INTENT_TRIANGLE'])[0].data
    return (verts, faces, img) if return_img else (verts, faces)


def write_gii(fname, verts, faces):
    # NOTE: SUMA only work with float32 NIFTI_INTENT_POINTSET and int32 NIFTI_INTENT_TRIANGLE
    verts = nibabel.gifti.GiftiDataArray(data=verts.astype('float32'), intent=nibabel.nifti1.intent_codes['NIFTI_INTENT_POINTSET'])
    faces = nibabel.gifti.GiftiDataArray(data=faces.astype('int32'), intent=nibabel.nifti1.intent_codes['NIFTI_INTENT_TRIANGLE'])
    img = nibabel.gifti.GiftiImage(darrays=[verts, faces])
    nibabel.gifti.write(img, fname)


def read_label(fname):
    '''Read FreeSurfer label'''
    x = read_txt(fname)
    nodes = np.int_(x[:,0])
    coords = x[:,1:4]
    labels = x[:,4]
    return nodes, coords, labels


# ========== NIML ascii ==========
NIML_DSET_CORE_TAGS = ['INDEX_LIST', 'SPARSE_DATA']

def read_niml_dset(fname, tags=None, as_asc=True, return_type='list'):
    if tags is None:
        tags = NIML_DSET_CORE_TAGS
    if as_asc:
        temp_file = 'tmp.' + fname
        if not path.exists(temp_file):
            subprocess.check_call(['ConvertDset', '-o_niml_asc', '-input', fname, '-prefix', temp_file])
        root = etree.parse(temp_file).getroot()
        os.remove(temp_file)
        def get_data(tag):
            element = root.find(tag)
            return np.fromiter(element.text.split(), dtype=element.get('ni_type'))
        data = {tag: get_data(tag) for tag in tags}
    if return_type == 'list':
        return [data[tag] for tag in tags]
    elif return_type == 'dict':
        return data
    elif return_type == 'tree':
        return root


# def read_niml_bin_nodes(fname):
#     '''
#     Read "Node Bucket" (node indices and values) from niml (binary) dataset.
#     This implementation is experimental for one-column dset only.
#     '''
#     with open(fname, 'rb') as fin:
#         s = fin.read()
#         data = []
#         for tag in NIML_DSET_CORE_TAGS:
#             pattern = '<{0}(.*?)>(.*?)</{0}>'.format(tag)
#             match = re.search(bytes(pattern, encoding='utf-8'), s, re.DOTALL)
#             if match is not None:
#                 # attrs = match.group(1).decode('utf-8').split()
#                 # attrs = {k: v[1:-1] for k, v in (attr.split('=') for attr in attrs)}
#                 attrs = shlex.split(match.group(1).decode('utf-8')) # Don't split quoted string
#                 attrs = dict(attr.split('=') for attr in attrs)
#                 x = np.frombuffer(match.group(2), dtype=attrs['ni_type']+'32')
#                 data.append(x.reshape(np.int_(attrs['ni_dimen'])))
#             else:
#                 data.append(None)
#         if data[0] is None: # Non-sparse dataset
#             data[0] = np.arange(data[1].shape[0])
#         return data[0], data[1]


# ========== NIML binary ==========
def read_niml_bin_nodes(fname):
    '''
    Read "Node Bucket" (node indices and values) from niml (binary) dataset.
    '''
    with open(fname, 'rb') as fin:
        s = fin.read()
        data = []
        for tag in NIML_DSET_CORE_TAGS:
            pattern = '<{0}(.*?)>(.*?)</{0}>'.format(tag)
            match = re.search(bytes(pattern, encoding='utf-8'), s, re.DOTALL)
            if match is not None:
                attrs = shlex.split(match.group(1).decode('utf-8')) # Don't split quoted string
                attrs = dict(attr.split('=') for attr in attrs)
                if '*' in attrs['ni_type']: # Multi-colume dataset
                    n, t = attrs['ni_type'].split('*')
                    attrs['n_columes'] = int(n)
                    attrs['dtype'] = t+'32'
                else:
                    attrs['n_columes'] = int(1)
                    attrs['dtype'] = attrs['ni_type']+'32'
                x = np.frombuffer(match.group(2), dtype=attrs['dtype'])
                data.append(x.reshape(np.int_([attrs['ni_dimen'], attrs['n_columes']])))
            else:
                data.append(None)
        if data[0] is None: # Non-sparse dataset
            data[0] = np.arange(data[1].shape[0])
        return data[0].squeeze(), data[1].squeeze()


# def write_niml_bin_nodes(fname, idx, val):
#     '''
#     Write "Node Bucket" (node indices and values) as niml (binary) dataset.
#     This implementation is experimental for one-column dset only.

#     References
#     ----------
#     [1] https://afni.nimh.nih.gov/afni/community/board/read.php?1,60396,60399#msg-60399
#     [2] After some trial-and-error, the following components are required:
#         self_idcode, COLMS_RANGE, COLMS_TYPE (tell suma how to interpret val), 
#         no whitespace between opening tag and binary data.
#     '''
#     with open(fname, 'wb') as fout:
#         # AFNI_dataset
#         fout.write('<AFNI_dataset dset_type="Node_Bucket" self_idcode="{0}" \
#             ni_form="ni_group">\n'.format(generate_afni_idcode()).encode('utf-8'))
#         # COLMS_RANGE
#         fout.write('<AFNI_atr ni_type="String" ni_dimen="1" atr_name="COLMS_RANGE">\
#             "{0} {1} {2} {3}"</AFNI_atr>\n'.format(np.min(val), np.max(val), 
#             idx[np.argmin(val)], idx[np.argmax(val)]).encode('utf-8'))
#         # COLMS_TYPE
#         col_types = {'int': 'Node_Index_Label', 'float': 'Generic_Float'}
#         fout.write('<AFNI_atr ni_type="String" ni_dimen="1" atr_name="COLMS_TYPE">\
#             "{0}"</AFNI_atr>\n'.format(col_types[get_ni_type(val)]).encode('utf-8'))
#         # INDEX_LIST
#         # Important: There should not be any \n after the opening tag for the binary data!
#         fout.write('<INDEX_LIST ni_form="binary.lsbfirst" ni_type="int" ni_dimen="{0}" \
#             data_type="Node_Bucket_node_indices">'.format(len(idx)).encode('utf-8'))
#         fout.write(idx.astype('int32').tobytes())
#         fout.write(b'</INDEX_LIST>\n')
#         # SPARSE_DATA
#         fout.write('<SPARSE_DATA ni_form="binary.lsbfirst" ni_type="{0}" ni_dimen="{1}" \
#             data_type="Node_Bucket_data">'.format(get_ni_type(val), len(val)).encode('utf-8'))
#         fout.write(val.astype(get_ni_type(val)+'32').tobytes())
#         fout.write(b'</SPARSE_DATA>\n')
#         fout.write(b'</AFNI_dataset>\n')


def write_niml_bin_nodes(fname, idx, val):
    '''
    Write "Node Bucket" (node indices and values) as niml (binary) dataset.

    References
    ----------
    [1] https://afni.nimh.nih.gov/afni/community/board/read.php?1,60396,60399#msg-60399
    [2] After some trial-and-error, the following components are required:
        self_idcode, COLMS_RANGE, COLMS_TYPE (tell suma how to interpret val), 
        no whitespace between opening tag and binary data.
    '''
    idx = np.atleast_1d(idx.squeeze())
    val = np.atleast_2d(val)
    if val.shape[0] == 1 and val.shape[1] == len(idx):
        val = val.T
    n_columes = val.shape[1]
    with open(fname, 'wb') as fout:
        # AFNI_dataset
        fout.write('<AFNI_dataset dset_type="Node_Bucket" self_idcode="{0}" \
            ni_form="ni_group">\n'.format(generate_afni_idcode()).encode('utf-8'))
        # COLMS_RANGE
        colms_range = ';'.join(['{0} {1} {2} {3}'.format(np.min(val[:,k]), np.max(val[:,k]), 
            idx[np.argmin(val[:,k])], idx[np.argmax(val[:,k])]) for k in range(n_columes)])
        fout.write('<AFNI_atr ni_type="String" ni_dimen="1" atr_name="COLMS_RANGE">\
            "{0}"</AFNI_atr>\n'.format(colms_range).encode('utf-8'))
        # COLMS_TYPE
        col_types = {'int': 'Node_Index_Label', 'float': 'Generic_Float'}
        colms_type = ';'.join(['{0}'.format(col_types[get_ni_type(val[:,k])]) for k in range(n_columes)])
        fout.write('<AFNI_atr ni_type="String" ni_dimen="1" atr_name="COLMS_TYPE">\
            "{0}"</AFNI_atr>\n'.format(colms_type).encode('utf-8'))
        # INDEX_LIST
        # Important: There should not be any \n after the opening tag for the binary data!
        fout.write('<INDEX_LIST ni_form="binary.lsbfirst" ni_type="int" ni_dimen="{0}" \
            data_type="Node_Bucket_node_indices">'.format(len(idx)).encode('utf-8'))
        fout.write(idx.astype('int32').tobytes())
        fout.write(b'</INDEX_LIST>\n')
        # SPARSE_DATA
        fout.write('<SPARSE_DATA ni_form="binary.lsbfirst" ni_type="{0}" ni_dimen="{1}" \
            data_type="Node_Bucket_data">'.format(get_ni_type(val), len(val)).encode('utf-8'))
        fout.write(val.astype(get_ni_type(val[:,0])+'32').tobytes())
        fout.write(b'</SPARSE_DATA>\n')
        fout.write(b'</AFNI_dataset>\n')


def generate_afni_idcode():
    return 'AFN_' + ''.join(random.choice(string.ascii_letters + string.digits) for n in range(22))


def get_ni_type(x):
    multiple = '{0}*'.format(x.shape[1]) if x.squeeze().ndim > 1 else ''
    if np.issubdtype(x.dtype, np.integer) or np.issubdtype(x.dtype, np.bool_):
        return multiple+'int'
    elif np.issubdtype(x.dtype, np.floating):
        return multiple+'float'


def write_1D_nodes(fname, idx, val):
    if idx is None:
        idx = np.arange(len(val))
    formats = dict(int='%d', float='%.6f')
    np.savetxt(fname, np.c_[idx, val], fmt=['%d', formats[get_ni_type(val)]])
    

# ========== Affine matrix (matvec format, or aff12) ==========
def read_affine(fname, sep=None):
    '''
    Returns
    -------
    mat : 3x4 or Nx3x4
    '''
    mat = read_txt(fname, delimiter=sep).reshape(-1,3,4).squeeze()
    return mat


def write_affine(fname, mat, oneline=True, sep=None):
    '''
    TODO: Not support multivolume affine yet
    '''
    if sep is None:
        sep = ' '
    with open(fname, 'w') as fo:
        if oneline:
            fo.write(sep.join(['%.6f' % x for x in mat.flat]) + '\n')
        else:
            for row in mat:
                fo.write(sep.join(['%.6f' % x for x in row]) + '\n')


def read_warp(fname):
    '''
    References
    ----------
    [1] https://afni.nimh.nih.gov/pub/dist/doc/program_help/3dQwarp.html
        "An AFNI nonlinear warp dataset stores the displacements (in DICOM mm) from
        the base dataset grid to the source dataset grid.
        AFNI stores a 3D warp as a 3-volume dataset (NiFTI or AFNI format), with the
        voxel values being the displacements in mm (32-bit floats) needed to
        'reach out' and bring (interpolate) another dataset into alignment -- that is,
        'pulling it back' to the grid defined in the warp dataset header."
    '''
    vol = read_vol(fname)
    dX, dY, dZ = np.rollaxis(vol.squeeze(), -1, 0)
    xyz2ijk = math.invert_affine(afni.get_affine(fname)) # iMAT
    return dX, dY, dZ, xyz2ijk


def read_register_dat(fname):
    mat = io.read_txt(fname, skiprows=4, nrows=3)
    return mat


class MaskDumper(object):
    def __init__(self, mask_file):
        self.mask_file = mask_file
        self.temp_file = 'tmp.dump.txt'
        subprocess.check_call(['3dmaskdump', '-mask', self.mask_file, '-index', '-xyz',
            '-o', self.temp_file, self.mask_file])
        x = np.loadtxt(self.temp_file)
        self.index = x[:,0].astype(int)
        self.ijk = x[:,1:4].astype(int)
        self.xyz = x[:,4:7]
        self.mask = x[:,7].astype(int)
        os.remove(self.temp_file)

    def dump(self, fname):
        files = glob.glob(fname) if isinstance(fname, six.string_types) else fname
        subprocess.check_call(['3dmaskdump', '-mask', self.mask_file, '-noijk',
            '-o', self.temp_file, ' '.join(files)])
        x = np.loadtxt(self.temp_file)
        os.remove(self.temp_file)
        return x

    def undump(self, prefix, x):
        np.savetxt(self.temp_file, np.c_[self.ijk, x])
        subprocess.check_call(['3dUndump', '-master', self.mask_file, '-ijk',
            '-prefix', prefix, '-overwrite', self.temp_file])
        os.remove(self.temp_file)


class Mask(object):
    def __init__(self, master=None, kind='mask'):
        self.master = master
        self.value = None
        if self.master is not None:
            self._infer_geometry(self.master)
            if master.endswith('.nii') or master.endswith('.nii.gz'):
                self.value = read_nii(self.master).ravel('F') # [x,y,z], x changes the fastest. Also, NIFTI read/write data in 'F'.
            else:
                self.value = read_afni(self.master).ravel('F')
            if kind == 'mask':
                idx = self.value > 0 # afni uses Fortran index here
                self.value = self.value[idx]
                self.index = np.arange(np.prod(self.IJK))[idx]
            elif kind == 'full':
                self.index = np.arange(np.prod(self.IJK))

    def _infer_geometry(self, fname):
        self.IJK = afni.get_DIMENSION(fname)[:3]
        self.MAT = afni.get_affine(fname)

    def to_dict(self):
        return dict(master=self.master, value=self.value, index=self.index, IJK=self.IJK, MAT=self.MAT)

    @classmethod
    def from_dict(cls, d):
        self = cls(None)
        for k, v in d.items():
            setattr(self, k, v)
        return self

    @classmethod
    def from_expr(cls, expr=None, **kwargs):
        master = list(kwargs.values())[0]
        mask = cls(master=None)
        mask.master = master
        mask._infer_geometry(master)
        data = {v: read_vol(f).squeeze() for v, f in kwargs.items()}
        idx = eval(expr, data).ravel('F') > 0
        mask.index = np.arange(np.prod(mask.IJK))[idx]
        return mask

    @classmethod
    def from_files(cls, files, combine='union'):
        mask = cls(files[0])
        mask.value[:] = 1
        for k in range(1, len(files)):
            mask2 = cls(files[k])
            mask2.value[:] = 2**k
            if combine == 'union':
                mask = mask + mask2
            elif combine == 'intersect':
                mask = mask * mask2
        return mask

    def compatible(self, other):
        return np.all(self.IJK==other.IJK) and np.allclose(self.MAT, other.MAT)

    def __repr__(self):
        return 'Mask ({0} voxels)'.format(len(self.index))

    def __add__(self, other):
        '''Mask union. Both masks are assumed to share the same grid.'''
        assert(self.compatible(other))
        mask = copy.deepcopy(self)
        mask.index = np.union1d(self.index, other.index)
        value_dict = {idx: val for idx, val in zip(self.index, self.value)}
        for idx, val in zip(other.index, other.value):
            if idx in value_dict:
                value_dict[idx] += val
            else:
                value_dict[idx] = val
        mask.value = np.array([value_dict[idx] for idx in mask.index])
        return mask

    def __mul__(self, other):
        '''Mask intersection. Both masks are assumed to share the same grid.'''
        assert(self.compatible(other))
        mask = copy.deepcopy(self)
        mask.index = np.intersect1d(self.index, other.index)
        return mask

    def __sub__(self, other):
        '''
        Voxels that are in the 1st mask but not in the 2nd mask.
        Both masks are assumed to share the same grid.
        '''
        assert(self.compatible(other))
        mask = copy.deepcopy(self)
        mask.index = mask.index[~np.in1d(self.index, other.index, assume_unique=True)]
        return mask

    def __contains__(self, other):
        assert(self.compatible(other))
        return np.all(np.in1d(other.index, self.index, assume_unique=True))

    def pick(self, selector, inplace=False):
        mask = self if inplace else copy.deepcopy(self)
        mask.index = mask.index[selector]
        return mask

    def constrain(self, func, return_selector=False, inplace=False):
        '''
        Parameters
        ----------
        func : callable
            selector = func(x, y, z) is used to select a subset of self.index
        '''
        ijk1 = np.c_[np.unravel_index(self.index, self.IJK, order='F') + (np.ones_like(self.index),)]
        xyz = np.dot(self.MAT, ijk1.T).T # Yes, it is xyz here!
        selector = func(xyz[:,0], xyz[:,1], xyz[:,2])
        mask = self if inplace else copy.deepcopy(self)
        mask.index = mask.index[selector]
        return mask if not return_selector else (mask, selector)

    def infer_selector(self, smaller):
        assert(smaller in self)
        selector = np.in1d(self.index, smaller.index, assume_unique=True)
        return selector

    def near(self, x, y, z, r, **kwargs):
        '''mm'''
        if np.isscalar(r):
            r = np.ones(3) * r
        func = (lambda X, Y, Z: ((X-x)/r[0])**2 + ((Y-y)/r[1])**2 + ((Z-z)/r[2])**2 < 1)
        return self.constrain(func, **kwargs)

    def ball(self, c, r, **kwargs):
        # return self.near(*c, r, **kwargs) # For python 2.7 compatibility
        return self.near(c[0], c[1], c[2], r, **kwargs)

    def cylinder(self, c, r, **kwargs):
        '''The elongated axis is represented as nan'''
        if np.isscalar(r):
            r = np.ones(3) * r
        func = (lambda X, Y, Z: np.nansum(np.c_[((X-c[0])/r[0])**2, ((Y-c[1])/r[1])**2, ((Z-c[2])/r[2])**2], axis=1) < 1)
        return self.constrain(func, **kwargs)

    def slab(self, x1=None, x2=None, y1=None, y2=None, z1=None, z2=None, **kwargs):
        limits = np.dot(self.MAT, np.c_[np.r_[0,0,0,1], np.r_[self.IJK-1,1]])
        x1 = np.min(limits[0,:]) if x1 is None else x1
        x2 = np.max(limits[0,:]) if x2 is None else x2
        y1 = np.min(limits[1,:]) if y1 is None else y1
        y2 = np.max(limits[1,:]) if y2 is None else y2
        z1 = np.min(limits[2,:]) if z1 is None else z1
        z2 = np.max(limits[2,:]) if z2 is None else z2
        func = (lambda X, Y, Z: (x1<X)&(X<x2) & (y1<Y)&(Y<y2) & (z1<Z)&(Z<z2))
        return self.constrain(func, **kwargs)

    def dump(self, fname, dtype=None):
        files = glob.glob(fname) if isinstance(fname, six.string_types) else fname
        # return np.vstack(read_afni(f).T.flat[self.index] for f in files).T.squeeze() # Cannot handle 4D...
        data = []
        for f in files:
            if f.endswith('.nii') or f.endswith('.nii.gz'):
                vol = read_nii(f)
            else:
                vol = read_afni(f)
            if dtype is not None:
                vol = vol.astype(dtype)
            S = vol.shape
            T = list(range(vol.ndim))
            T[:3] = T[:3][::-1]
            data.append(vol.transpose(*T).reshape(np.prod(S[:3]),int(np.prod(S[3:])))[self.index,:])
        return np.hstack(data).squeeze()

    def undump(self, prefix, x, method='nibabel', space=None):
        if method == 'nibabel': # Much faster
            temp_file = 'tmp.%s.nii' % next(tempfile._get_candidate_names())
            vol = np.zeros(self.IJK) # Don't support int64?？
            assert(self.index.size==x.size)
            vol.T.flat[self.index] = x
            mat = np.dot(np.diag([-1,-1, 1]), self.MAT) # AFNI uses DICOM's RAI, but NIFTI uses LPI aka RAS+
            aff = nibabel.affines.from_matvec(mat[:,:3], mat[:,3])
            img = nibabel.Nifti1Image(vol, aff)
            # https://afni.nimh.nih.gov/afni/community/board/read.php?1,149338,149340#msg-149340
            # 0 (unknown) sform not defined
            # 1 (scanner) RAS+ in scanner coordinates
            # 2 (aligned) RAS+ aligned to some other scan
            # 3 (talairach) RAS+ in Talairach atlas space
            # 4 (mni) RAS+ in MNI atlas space
            if space is None:
                space = 1
            img.header['sform_code'] = space
            if prefix.endswith('.nii'):
                nibabel.save(img, prefix)
            else:
                nibabel.save(img, temp_file)
                subprocess.check_call(['3dcopy', temp_file, prefix+'+orig', '-overwrite']) # However, still TLRC inside...
                os.remove(temp_file)
        elif method == '3dUndump': # More robust
            temp_file = 'tmp.%s.txt' % next(tempfile._get_candidate_names())
            ijk = np.c_[np.unravel_index(self.index, self.IJK, order='F')]
            np.savetxt(temp_file, np.c_[ijk, x])
            subprocess.check_call(['3dUndump', '-master', self.master, '-ijk',
                '-prefix', prefix, '-overwrite', temp_file])
            os.remove(temp_file)

    @property
    def ijk(self):
        return np.c_[np.unravel_index(self.index, self.IJK, order='F')]

    @property
    def xyz(self):
        return np.dot(self.MAT[:,:3], self.ijk.T).T + self.MAT[:,3]

    @property
    def xyz_nifti(self):
        return self.xyz * np.r_[-1,-1,1] # AFNI uses DICOM's RAI, but NIFTI uses LPI aka RAS+


class BallMask(Mask):
    def __init__(self, master, c, r):
        Mask.__init__(self, master, kind='full')
        self.ball(c, r, inplace=True)


class CylinderMask(Mask):
    def __init__(self, master, c, r):
        Mask.__init__(self, master, kind='full')
        self.cylinder(c, r, inplace=True)


class SlabMask(Mask):
    def __init__(self, master, x1=None, x2=None, y1=None, y2=None, z1=None, z2=None):
        Mask.__init__(self, master, kind='full')
        self.slab(x1, x2, y1, y2, z1, z2, inplace=True)


def filter_cluster(in_file, out_file, top=None, neighbor=2):
    '''
    neighbor : int
        1 : face touch
        2 : edge touch (default, as in afni)
        3 : corner touch
    '''
    d = np.mgrid[-1:2,-1:2,-1:2]
    structure = (np.linalg.norm(d, axis=0) <= (neighbor+1)/2).astype(int)
    im, img = read_vol(in_file, return_img=True)
    label, n = ndimage.label(im, structure)
    vol = [np.sum(label==k) for k in range(1, n+1)]
    if top is not None:
        keeped = np.argsort(vol)[::-1][:top] + 1
        for k in range(1, n+1):
            if k not in keeped:
                im[label==k] = 0
    write_vol(out_file, im, base_img=img)


if __name__ == '__main__':
    pass
